from logging import Logger

from typing import (
    List, Optional,
)

from .._OrderChangerProductionBox import OrderChangerProductionBox
from production.model import ProductionOrder


class PrepTimeOrderSplitter(OrderChangerProductionBox):
    def __init__(
        self,                   # type: PrepTimeOrderSplitter
        name,                   # type: str
        sons,                   # type: object
        logger=None,            # type: Optional[Logger]
    ):
        # type: (...) -> None

        super(PrepTimeOrderSplitter, self).__init__(name, sons, logger)

    def change_order(
        self,       # type: PrepTimeOrderSplitter
        order,      # type: ProductionOrder
    ):
        # type: (...) -> List[ProductionOrder]

        orders = []
        split_index = 0

        for item in order.items:
            for temp_order in orders:
                if len(temp_order.items) > 0 and item.prep_time == temp_order.items[0].prep_time:
                    temp_order.items.append(item)
                    break
            else:
                new_order = order.clone([item], split_index)
                orders.append(new_order)
                split_index += 1

        return orders



from typing import (
    Any,
    Dict,
    List,
)
from xml.etree import cElementTree as eTree

from common import (
    FiscalOrderTotals,
    MobileProductParser,
)
from nreports_api.shared import JSON
from .._BaseFiscalNfce import BaseFiscalNfce


class BaseFiscalCouponMobileNfce(BaseFiscalNfce):

    def build_data(
        self,
        data_to_build: Dict[str, JSON],
        **kwargs: Any,
    ) -> Dict[str, JSON]:

        fiscal_xml = kwargs["fiscalXml"]

        fiscal_order_totals = FiscalOrderTotals(fiscal_xml=fiscal_xml)

        self._build_header(
            data_to_build=data_to_build,
            fiscal_xml=fiscal_xml,
        )

        self._build_body(
            data_to_build=data_to_build,
            fiscal_xml=fiscal_xml,
            fiscal_order_totals=fiscal_order_totals,
            products=self._get_item_products(fiscal_xml=fiscal_xml),
        )

        self._build_footer(
            data_to_build=data_to_build,
            fiscal_xml=fiscal_xml,
        )

        return data_to_build

    @staticmethod
    def _get_item_products(
        fiscal_xml: eTree.Element,
    ) -> List[Dict[str, str]]:

        item_products: List[Dict[str, str]] = []
        fiscal_products = fiscal_xml.findall(".//prod")
        product_params = ["cProd", "xProd", "qCom", "uCom", "vUnCom", "vProd"]
        for product in fiscal_products:
            product_item = MobileProductParser(
                product_xml=product,
                params=product_params,
            )

            item_products.append(
                {
                    "ITEM_PRODUCT_CODE": product_item.product_code,
                    "ITEM_PRODUCT_NAME": product_item.product_name,
                    "ITEM_PRODUCT_QUANTITY": product_item.product_qty,
                    "ITEM_PRODUCT_UNIT": product_item.product_unit.upper(),
                    "ITEM_PRODUCT_PRICE": product_item.product_price,
                    "ITEM_PRODUCT_TOTAL_PRICE": product_item.product_total_price,
                }
            )

        return item_products



from ._BaseFiscalCouponMobileNfce import BaseFiscalCouponMobileNfce
from ._BaseFiscalCouponMobileSat import BaseFiscalCouponMobileSat
from ._FiscalCouponMobileNfceLargeBuilder import FiscalCouponMobileNfceLargeBuilder
from ._FiscalCouponMobileNfceSmallBuilder import FiscalCouponMobileNfceSmallBuilder
from ._FiscalCouponMobileSatLargeBuilder import FiscalCouponMobileSatLargeBuilder
from ._FiscalCouponMobileSatSmallBuilder import FiscalCouponMobileSatSmallBuilder



from helper import remove_accents
from pos_model import Order, XmlTenderTypeFromPOSMapper, TenderType, OrderTender

from _SatXmlPartBuilder import SatXmlPartBuilder
from satbuilder import ContextKeys
from typing import Dict


class PgtoBuilder(SatXmlPartBuilder):
    def __init__(self, tenders_type):
        # type: (Dict) -> None

        self.tenders_type = tenders_type
        self.tenders = self._get_tenders()

    def _get_tenders(
        self
    ):
        # type: (...) -> Dict[int, Dict]

        tenders = {}
        for tender in self.tenders_type.values():
            tenders[int(tender["tenderId"])] = tender

        return tenders

    def build_xml(self, order, context):
        # type: (Order, ContextKeys) -> unicode

        xml = "<pgto>"
        tag = "<MP><cMP>{0:>02}</cMP><vMP>{1:>0.2f}</vMP><cAdmC>{2}</cAdmC></MP>"

        max_order_tenders = 10

        if len(order.tenders) <= max_order_tenders:
            for tender in order.tenders:
                tender_type = XmlTenderTypeFromPOSMapper.get(tender.type)
                if tender.type in self.tenders:
                    t_pag = self.tenders[tender.type].get("tPag")
                    if t_pag is not None:
                        tender_type = int(self.tenders[tender.type].get("tPag"))

                cad_mc = self._get_cad_mc(tender)
                xml += tag.format(tender_type, tender.amount, cad_mc)
        else:
            tenders_amount = {}
            for tender in order.tenders:
                cad_mc = self._get_cad_mc(tender)
                tender_type = XmlTenderTypeFromPOSMapper.get(tender.type)

                if cad_mc not in tenders_amount:
                    tenders_amount[cad_mc] = {}
                if tender_type not in tenders_amount[cad_mc]:
                    tenders_amount[cad_mc][tender_type] = 0

                tenders_amount[cad_mc][tender_type] += tender.amount

            payment_groups = 0
            for cad_mc, tenders in tenders_amount.items():
                for tender_type in tenders:
                    amount = tenders[tender_type]
                    xml += tag.format(tender_type, amount, cad_mc)
                    payment_groups += 1

            if payment_groups > max_order_tenders:
                raise Exception(
                    "$ORDER_NOT_FINALIZED_DUE_TO_MAX_SAT_PAYMENT_GROUPS_REACHED|{}".format(max_order_tenders)
                )

        xml += "</pgto>"

        return xml

    def _get_cad_mc(
            self,
            tender  # type: OrderTender
    ):
        # type: (...) -> str

        if tender.type in [TenderType.credit, TenderType.debit] and tender.accreditor_sat not in ['', None]:
            return str(tender.accreditor_sat)

        others = "999"
        tender_description = str(remove_accents(tender.descr)).upper()
        if tender_description in self.tenders_type:
            tender_type_data_base = self.tenders_type[tender_description]
            cad_mc = str(tender_type_data_base.get("cAdmC", ""))
            return cad_mc if cad_mc.strip() else others

        return others



from xml.etree import cElementTree as eTree


def order_was_already_saved(
    order,  # type: eTree.XML
):
    # type: (...) -> bool

    state_history = order.findall(".//StateHistory/State")
    for state in state_history:
        if state.get("state") == "STORED":
            return True

    return False



import sys


def disable_importlib_metadata_finder(metadata):
    """
    Ensure importlib_metadata doesn't provide older, incompatible
    Distributions.

    Workaround for #3102.
    """
    try:
        import importlib_metadata
    except ImportError:
        return
    except AttributeError:
        from .warnings import SetuptoolsWarning

        SetuptoolsWarning.emit(
            "Incompatibility problem.",
            """
            `importlib-metadata` version is incompatible with `setuptools`.
            This problem is likely to be solved by installing an updated version of
            `importlib-metadata`.
            """,
            see_url="https://github.com/python/importlib_metadata/issues/396",
        )  # Ensure a descriptive message is shown.
        raise  # This exception can be suppressed by _distutils_hack

    if importlib_metadata is metadata:
        return
    to_remove = [
        ob
        for ob in sys.meta_path
        if isinstance(ob, importlib_metadata.MetadataPathFinder)
    ]
    for item in to_remove:
        sys.meta_path.remove(item)


if sys.version_info < (3, 10):
    from setuptools.extern import importlib_metadata as metadata

    disable_importlib_metadata_finder(metadata)
else:
    import importlib.metadata as metadata


if sys.version_info < (3, 9):
    from setuptools.extern import importlib_resources as resources
else:
    import importlib.resources as resources  # noqa: F401



# eic.py - functions for handling EU EIC numbers
# coding: utf-8
#
# Copyright (C) 2017 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""EIC (European Energy Identification Code).

The EIC (Energy Identification Code) a 16 character code used in Europe to
uniquely identify entities and objects in the electricity and gas sector.

The number uses letters, digits and the minus sign. The first 2 character
identify the issuing office, 1 character for the object type, 12 digits for
the object and 1 check character.

More information:

* https://en.wikipedia.org/wiki/Energy_Identification_Code
* https://www.entsoe.eu/data/energy-identification-codes-eic/

>>> validate('22XWATTPLUS----G')
'22XWATTPLUS----G'
>>> validate('22XWATTPLUS----X')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('23X--130302DLGW-')  # check digit cannot be minus
Traceback (most recent call last):
    ...
InvalidFormat: ...
"""

from stdnum.exceptions import *
from stdnum.util import clean


_alphabet = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ-'


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding white space."""
    return clean(number, ' ').strip()


def calc_check_digit(number):
    """Calculate the check digit for the number."""
    number = compact(number)
    s = sum((16 - i) * _alphabet.index(n) for i, n in enumerate(number[:15]))
    return _alphabet[36 - ((s - 1) % 37)]


def validate(number):
    """Check if the number is valid. This checks the length, format and check
    digit."""
    number = compact(number)
    if not all(x in _alphabet for x in number):
        raise InvalidFormat()
    if len(number) != 16:
        raise InvalidLength()
    if number[-1] == '-':
        raise InvalidFormat()
    if number[-1] != calc_check_digit(number):
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is valid. This checks the length, format and check
    digit."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



from production.box._OrderChangerProductionBox import OrderChangerProductionBox


class PointerBox(OrderChangerProductionBox):
    def __init__(self, name, sons, points, default_points, count_merged_items, logger=None):
        super(PointerBox, self).__init__(name, sons, logger)
        self.points = points
        self.default_points = default_points
        self.count_merged_items = count_merged_items

    def change_order(self, order):
        order.points = self.add_points(order.items)
        return order

    def add_points(self, items):
        order_points = 0
        for item in items:
            if not self.count_merged_items and hasattr(item, "stamped"):
                continue
            if item.is_product():
                if item.part_code in self.points:
                    item.points = self.points[item.part_code] * item.multiplied_qty
                else:
                    item.points = self.default_points * item.multiplied_qty
                order_points += item.points
            else:
                self.add_points(item.items)
        return order_points



from messagebus import TokenCreator, TokenPriority


MSGGRP_SLCTRL_ORDERMGR = "A"

TK_SLCTRL_OMGR_DOSALE = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="1",
)

TK_SLCTRL_OMGR_DOOPTION = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="A",
)

TK_SLCTRL_OMGR_ORDERPICT = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="2"
)

TK_SLCTRL_OMGR_CREATEORDER = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="18"
)

TK_SLCTRL_OMGR_DOTENDER = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="10",
)

TK_SLCTRL_OMGR_LISTORDERS = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="D"
)

TK_SLCTRL_OMGR_CHANGEQTY = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="E",
)

TK_SLCTRL_OMGR_ORDERREOPEN = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="F",
)

TK_SLCTRL_OMGR_ORDERTOTAL = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="6",
)

TK_SLCTRL_OMGR_PULLORDERS = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="34",
)

TK_SLCTRL_OMGR_RECALLORDER = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="C",
)

TK_SLCTRL_OMGR_STOREORDER = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="B",
)

TK_SLCTRL_OMGR_VOIDLINE = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="3",
)
TK_SLCTRL_OMGR_VOIDORDER = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="4",
)

TK_SLCTRL_OMGR_CLEAROPTION = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="7",
)

TK_SLCTRL_OMGR_DOOPTION = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="A",
)

TK_SLCTRL_OMGR_CHANGEMODIFIER = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_ORDERMGR,
    number="13",
)


def get_tokens():
    return (
        TK_SLCTRL_OMGR_CHANGEMODIFIER,
        TK_SLCTRL_OMGR_CHANGEQTY,
        TK_SLCTRL_OMGR_CLEAROPTION,
        TK_SLCTRL_OMGR_DOOPTION,
        TK_SLCTRL_OMGR_DOSALE,
        TK_SLCTRL_OMGR_DOTENDER,
        TK_SLCTRL_OMGR_ORDERPICT,
        TK_SLCTRL_OMGR_CREATEORDER,
        TK_SLCTRL_OMGR_LISTORDERS,
        TK_SLCTRL_OMGR_ORDERREOPEN,
        TK_SLCTRL_OMGR_ORDERTOTAL,
        TK_SLCTRL_OMGR_PULLORDERS,
        TK_SLCTRL_OMGR_RECALLORDER,
        TK_SLCTRL_OMGR_STOREORDER,
        TK_SLCTRL_OMGR_VOIDLINE,
        TK_SLCTRL_OMGR_VOIDORDER,
    )



# Copyright 2020 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementation of the metadata abstraction for gRPC Asyncio Python."""
from collections import OrderedDict
from collections import abc
from typing import Any, Iterator, List, Tuple, Union

MetadataKey = str
MetadataValue = Union[str, bytes]


class Metadata(abc.Mapping):
    """Metadata abstraction for the asynchronous calls and interceptors.

    The metadata is a mapping from str -> List[str]

    Traits
        * Multiple entries are allowed for the same key
        * The order of the values by key is preserved
        * Getting by an element by key, retrieves the first mapped value
        * Supports an immutable view of the data
        * Allows partial mutation on the data without recreating the new object from scratch.
    """

    def __init__(self, *args: Tuple[MetadataKey, MetadataValue]) -> None:
        self._metadata = OrderedDict()
        for md_key, md_value in args:
            self.add(md_key, md_value)

    @classmethod
    def from_tuple(cls, raw_metadata: tuple):
        if raw_metadata:
            return cls(*raw_metadata)
        return cls()

    def add(self, key: MetadataKey, value: MetadataValue) -> None:
        self._metadata.setdefault(key, [])
        self._metadata[key].append(value)

    def __len__(self) -> int:
        """Return the total number of elements that there are in the metadata,
        including multiple values for the same key.
        """
        return sum(map(len, self._metadata.values()))

    def __getitem__(self, key: MetadataKey) -> MetadataValue:
        """When calling <metadata>[<key>], the first element of all those
        mapped for <key> is returned.
        """
        try:
            return self._metadata[key][0]
        except (ValueError, IndexError) as e:
            raise KeyError("{0!r}".format(key)) from e

    def __setitem__(self, key: MetadataKey, value: MetadataValue) -> None:
        """Calling metadata[<key>] = <value>
        Maps <value> to the first instance of <key>.
        """
        if key not in self:
            self._metadata[key] = [value]
        else:
            current_values = self.get_all(key)
            self._metadata[key] = [value, *current_values[1:]]

    def __delitem__(self, key: MetadataKey) -> None:
        """``del metadata[<key>]`` deletes the first mapping for <key>."""
        current_values = self.get_all(key)
        if not current_values:
            raise KeyError(repr(key))
        self._metadata[key] = current_values[1:]

    def delete_all(self, key: MetadataKey) -> None:
        """Delete all mappings for <key>."""
        del self._metadata[key]

    def __iter__(self) -> Iterator[Tuple[MetadataKey, MetadataValue]]:
        for key, values in self._metadata.items():
            for value in values:
                yield (key, value)

    def get_all(self, key: MetadataKey) -> List[MetadataValue]:
        """For compatibility with other Metadata abstraction objects (like in Java),
        this would return all items under the desired <key>.
        """
        return self._metadata.get(key, [])

    def set_all(self, key: MetadataKey, values: List[MetadataValue]) -> None:
        self._metadata[key] = values

    def __contains__(self, key: MetadataKey) -> bool:
        return key in self._metadata

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            return self._metadata == other._metadata
        if isinstance(other, tuple):
            return tuple(self) == other
        return NotImplemented  # pytype: disable=bad-return-type

    def __add__(self, other: Any) -> "Metadata":
        if isinstance(other, self.__class__):
            return Metadata(*(tuple(self) + tuple(other)))
        if isinstance(other, tuple):
            return Metadata(*(tuple(self) + other))
        return NotImplemented  # pytype: disable=bad-return-type

    def __repr__(self) -> str:
        view = tuple(self)
        return "{0}({1!r})".format(self.__class__.__name__, view)



# damm.py - functions for performing the Damm checksum algorithm
#
# Copyright (C) 2016-2021 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""The Damm algorithm.

The Damm algorithm is a check digit algorithm that should detect all
single-digit errors and all adjacent transposition errors. Based on
anti-symmetric quasigroup of order 10 it uses a substitution table.

This implementation uses the table from Wikipedia by default but a custom
table can be provided.

More information:

* https://en.wikipedia.org/wiki/Damm_algorithm

>>> validate('572')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> calc_check_digit('572')
'4'
>>> validate('5724')
'5724'
>>> table = (
...     (0, 2, 3, 4, 5, 6, 7, 8, 9, 1),
...     (2, 0, 4, 1, 7, 9, 5, 3, 8, 6),
...     (3, 7, 0, 5, 2, 8, 1, 6, 4, 9),
...     (4, 1, 8, 0, 6, 3, 9, 2, 7, 5),
...     (5, 6, 2, 9, 0, 7, 4, 1, 3, 8),
...     (6, 9, 7, 3, 1, 0, 8, 5, 2, 4),
...     (7, 5, 1, 8, 4, 2, 0, 9, 6, 3),
...     (8, 4, 6, 2, 9, 5, 3, 0, 1, 7),
...     (9, 8, 5, 7, 3, 1, 6, 4, 0, 2),
...     (1, 3, 9, 6, 8, 4, 2, 7, 5, 0))
>>> checksum('816', table=table)
9
"""

from stdnum.exceptions import *


_operation_table = (
    (0, 3, 1, 7, 5, 9, 8, 6, 4, 2),
    (7, 0, 9, 2, 1, 5, 4, 8, 6, 3),
    (4, 2, 0, 6, 8, 7, 1, 3, 5, 9),
    (1, 7, 5, 0, 9, 8, 3, 4, 2, 6),
    (6, 1, 2, 3, 0, 4, 5, 9, 7, 8),
    (3, 6, 7, 4, 2, 0, 9, 5, 8, 1),
    (5, 8, 6, 9, 7, 2, 0, 1, 3, 4),
    (8, 9, 4, 5, 3, 6, 2, 0, 1, 7),
    (9, 4, 3, 8, 6, 1, 7, 2, 0, 5),
    (2, 5, 8, 1, 4, 3, 6, 7, 9, 0))


def checksum(number, table=None):
    """Calculate the Damm checksum over the provided number. The checksum is
    returned as an integer value and should be 0 when valid."""
    table = table or _operation_table
    i = 0
    for n in str(number):
        i = table[i][int(n)]
    return i


def validate(number, table=None):
    """Check if the number provided passes the Damm algorithm."""
    if not bool(number):
        raise InvalidFormat()
    try:
        valid = checksum(number, table=table) == 0
    except Exception:  # noqa: B902
        raise InvalidFormat()
    if not valid:
        raise InvalidChecksum()
    return number


def is_valid(number, table=None):
    """Check if the number provided passes the Damm algorithm."""
    try:
        return bool(validate(number, table=table))
    except ValidationError:
        return False


def calc_check_digit(number, table=None):
    """Calculate the extra digit that should be appended to the number to
    make it a valid number."""
    return str(checksum(number, table=table))



from production.box import SaleTypeFilterBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class SaleTypeFilterBoxBuilder(BoxBuilder):
    def __init__(self, loggers):
        super(SaleTypeFilterBoxBuilder, self).__init__(loggers)

    def build(
        self,       # type: SaleTypeFilterBoxBuilder
        config,     # type: BoxConfiguration
    ):
        # type: (...) -> SaleTypeFilterBox

        return SaleTypeFilterBox(
            name=config.name,
            sons=config.sons,
            sale_types=self.get_extra(config, "AllowedSaleTypes", None),
            logger=self.get_logger(config),
        )



from xml.etree import cElementTree as eTree

from sysactions import (
    get_model,
    has_current_order,
    get_current_order,
)

from actions.order.order_was_already_saved import order_was_already_saved
from actions.login.pause_user import pause_user
from actions.table.store_table_service import store_table_service
from api.order import get_order_service
from .. import logger


def close_active_order_and_pause_user(
    pos_id,         # type: str
    void_reason,    # type: str
):
    # type: (...) -> None

    try:
        model = get_model(posid=pos_id)
        if has_current_order(model=model):
            _close_order(
                model=model,
                pos_id=pos_id,
                void_reason=void_reason,
            )

        logger.debug("Call pause user")
        pause_user(
            pos_id=pos_id,
        )
    except Exception as ex:
        logger.exception("Close active order because", exc_info=ex)


def _close_order(
    model,          # type: eTree.XML
    pos_id,         # type: str
    void_reason,    # type: str
):
    # type: (...) -> None

    current_order = get_current_order(model=model)
    table_id = current_order.find("CustomOrderProperties/OrderProperty/[@key='TABLE_ID']")
    if table_id is not None:
        _close_table_service(
            pos_id=pos_id,
            order=current_order,
            received_void_reason=void_reason,
        )
    else:
        _close_quick_service_order(
            model=model,
            pos_id=pos_id,
            order=current_order,
            received_void_reason=void_reason,
        )


def _close_quick_service_order(
    model,                  # type: eTree.XML
    pos_id,                 # type: str
    order,                  # type: eTree.Element
    received_void_reason,   # type: str
):
    # type: (...) -> None

    logger.debug("Start close quick service order")
    if order.get("state") not in ("IN_PROGRESS", "TOTALED"):
        return

    if order_was_already_saved(order=order):
        logger.debug("Store order was already saved")
        order_service = get_order_service(pos_id=pos_id)
        order_service.store_order()
        return

    logger.debug("Do void without authorization")
    from actions.void.do_void_order import do_void_without_authorization
    do_void_without_authorization(
        pos_id=pos_id,
        received_void_reason=received_void_reason,
    )


def _close_table_service(
    pos_id,                 # type: str
    order,                  # type: eTree.Element
    received_void_reason,   # type: str
):
    # type: (...) -> None

    logger.debug("Start close table service")
    if order.get("state") == "IN_PROGRESS" and not order_was_already_saved(order=order):
        logger.debug("Do void without authorization")
        from actions.void.do_void_order import do_void_without_authorization
        do_void_without_authorization(
            pos_id=pos_id,
            received_void_reason=received_void_reason,
        )

    logger.debug("Store table service")
    store_table_service(
        pos_id=pos_id,
    )



# coding=UTF-8
from __future__ import absolute_import

import copy
import datetime
import pickle

import hypothesis
import hypothesis.extra.pytz
import hypothesis.strategies
import pytest

from . import iso8601


def test_iso8601_regex() -> None:
    assert iso8601.ISO8601_REGEX.match("2006-10-11T00:14:33Z")


def test_fixedoffset_eq() -> None:
    # See https://bitbucket.org/micktwomey/pyiso8601/issues/19
    expected_timezone = datetime.timezone(offset=datetime.timedelta(hours=2))
    assert expected_timezone == iso8601.FixedOffset(2, 0, "+2:00")


def test_parse_no_timezone_different_default() -> None:
    tz = iso8601.FixedOffset(2, 0, "test offset")
    d = iso8601.parse_date("2007-01-01T08:00:00", default_timezone=tz)
    assert d == datetime.datetime(2007, 1, 1, 8, 0, 0, 0, tz)
    assert d.tzinfo == tz


def test_parse_utc_different_default() -> None:
    """Z should mean 'UTC', not 'default'."""
    tz = iso8601.FixedOffset(2, 0, "test offset")
    d = iso8601.parse_date("2007-01-01T08:00:00Z", default_timezone=tz)
    assert d == datetime.datetime(2007, 1, 1, 8, 0, 0, 0, iso8601.UTC)


@pytest.mark.parametrize(
    "invalid_date, error_string",
    [
        ("2013-10-", "Unable to parse date string"),
        ("2013-", "Unable to parse date string"),
        ("", "Unable to parse date string"),
        ("wibble", "Unable to parse date string"),
        ("23", "Unable to parse date string"),
        ("131015T142533Z", "Unable to parse date string"),
        ("131015", "Unable to parse date string"),
        ("20141", "Unable to parse date string"),
        ("201402", "Unable to parse date string"),
        (
            "2007-06-23X06:40:34.00Z",
            "Unable to parse date string",
        ),  # https://code.google.com/p/pyiso8601/issues/detail?id=14
        (
            "2007-06-23 06:40:34.00Zrubbish",
            "Unable to parse date string",
        ),  # https://code.google.com/p/pyiso8601/issues/detail?id=14
        ("20114-01-03T01:45:49", "Unable to parse date string"),
    ],
)
def test_parse_invalid_date(invalid_date: str, error_string: str) -> None:
    assert iso8601.is_iso8601(invalid_date) is False
    with pytest.raises(iso8601.ParseError) as exc:
        iso8601.parse_date(invalid_date)
    assert exc.errisinstance(iso8601.ParseError)
    assert str(exc.value).startswith(error_string)


@pytest.mark.parametrize(
    "valid_date,expected_datetime,isoformat",
    [
        (
            "2007-06-23 06:40:34.00Z",
            datetime.datetime(2007, 6, 23, 6, 40, 34, 0, iso8601.UTC),
            "2007-06-23T06:40:34+00:00",
        ),  # Handle a separator other than T
        (
            "1997-07-16T19:20+01:00",
            datetime.datetime(
                1997, 7, 16, 19, 20, 0, 0, iso8601.FixedOffset(1, 0, "+01:00")
            ),
            "1997-07-16T19:20:00+01:00",
        ),  # Parse with no seconds
        (
            "2007-01-01T08:00:00",
            datetime.datetime(2007, 1, 1, 8, 0, 0, 0, iso8601.UTC),
            "2007-01-01T08:00:00+00:00",
        ),  # Handle timezone-less dates. Assumes UTC. http://code.google.com/p/pyiso8601/issues/detail?id=4
        (
            "2006-10-20T15:34:56.123+02:30",
            datetime.datetime(
                2006, 10, 20, 15, 34, 56, 123000, iso8601.FixedOffset(2, 30, "+02:30")
            ),
            None,
        ),
        (
            "2006-10-20T15:34:56Z",
            datetime.datetime(2006, 10, 20, 15, 34, 56, 0, iso8601.UTC),
            "2006-10-20T15:34:56+00:00",
        ),
        (
            "2007-5-7T11:43:55.328Z",
            datetime.datetime(2007, 5, 7, 11, 43, 55, 328000, iso8601.UTC),
            "2007-05-07T11:43:55.328000+00:00",
        ),  # http://code.google.com/p/pyiso8601/issues/detail?id=6
        (
            "2006-10-20T15:34:56.123Z",
            datetime.datetime(2006, 10, 20, 15, 34, 56, 123000, iso8601.UTC),
            "2006-10-20T15:34:56.123000+00:00",
        ),
        (
            "2013-10-15T18:30Z",
            datetime.datetime(2013, 10, 15, 18, 30, 0, 0, iso8601.UTC),
            "2013-10-15T18:30:00+00:00",
        ),
        (
            "2013-10-15T22:30+04",
            datetime.datetime(
                2013, 10, 15, 22, 30, 0, 0, iso8601.FixedOffset(4, 0, "+04:00")
            ),
            "2013-10-15T22:30:00+04:00",
        ),  # <time>±hh:mm
        (
            "2013-10-15T1130-0700",
            datetime.datetime(
                2013, 10, 15, 11, 30, 0, 0, iso8601.FixedOffset(-7, 0, "-07:00")
            ),
            "2013-10-15T11:30:00-07:00",
        ),  # <time>±hhmm
        (
            "2013-10-15T1130+0700",
            datetime.datetime(
                2013, 10, 15, 11, 30, 0, 0, iso8601.FixedOffset(+7, 0, "+07:00")
            ),
            "2013-10-15T11:30:00+07:00",
        ),  # <time>±hhmm
        (
            "2013-10-15T1130+07",
            datetime.datetime(
                2013, 10, 15, 11, 30, 0, 0, iso8601.FixedOffset(+7, 0, "+07:00")
            ),
            "2013-10-15T11:30:00+07:00",
        ),  # <time>±hh
        (
            "2013-10-15T1130-07",
            datetime.datetime(
                2013, 10, 15, 11, 30, 0, 0, iso8601.FixedOffset(-7, 0, "-07:00")
            ),
            "2013-10-15T11:30:00-07:00",
        ),  # <time>±hh
        (
            "2013-10-15T15:00-03:30",
            datetime.datetime(
                2013, 10, 15, 15, 0, 0, 0, iso8601.FixedOffset(-3, -30, "-03:30")
            ),
            "2013-10-15T15:00:00-03:30",
        ),
        (
            "2013-10-15T183123Z",
            datetime.datetime(2013, 10, 15, 18, 31, 23, 0, iso8601.UTC),
            "2013-10-15T18:31:23+00:00",
        ),  # hhmmss
        (
            "2013-10-15T1831Z",
            datetime.datetime(2013, 10, 15, 18, 31, 0, 0, iso8601.UTC),
            "2013-10-15T18:31:00+00:00",
        ),  # hhmm
        (
            "2013-10-15T18Z",
            datetime.datetime(2013, 10, 15, 18, 0, 0, 0, iso8601.UTC),
            "2013-10-15T18:00:00+00:00",
        ),  # hh
        (
            "2013-10-15",
            datetime.datetime(2013, 10, 15, 0, 0, 0, 0, iso8601.UTC),
            "2013-10-15T00:00:00+00:00",
        ),  # YYYY-MM-DD
        (
            "20131015T18:30Z",
            datetime.datetime(2013, 10, 15, 18, 30, 0, 0, iso8601.UTC),
            "2013-10-15T18:30:00+00:00",
        ),  # YYYYMMDD
        (
            "2012-12-19T23:21:28.512400+00:00",
            datetime.datetime(
                2012, 12, 19, 23, 21, 28, 512400, iso8601.FixedOffset(0, 0, "+00:00")
            ),
            "2012-12-19T23:21:28.512400+00:00",
        ),  # https://code.google.com/p/pyiso8601/issues/detail?id=21
        (
            "2006-10-20T15:34:56.123+0230",
            datetime.datetime(
                2006, 10, 20, 15, 34, 56, 123000, iso8601.FixedOffset(2, 30, "+02:30")
            ),
            "2006-10-20T15:34:56.123000+02:30",
        ),  # https://code.google.com/p/pyiso8601/issues/detail?id=18
        (
            "19950204",
            datetime.datetime(1995, 2, 4, tzinfo=iso8601.UTC),
            "1995-02-04T00:00:00+00:00",
        ),  # https://code.google.com/p/pyiso8601/issues/detail?id=1
        (
            "2010-07-20 15:25:52.520701+00:00",
            datetime.datetime(
                2010, 7, 20, 15, 25, 52, 520701, iso8601.FixedOffset(0, 0, "+00:00")
            ),
            "2010-07-20T15:25:52.520701+00:00",
        ),  # https://code.google.com/p/pyiso8601/issues/detail?id=17
        (
            "2010-06-12",
            datetime.datetime(2010, 6, 12, tzinfo=iso8601.UTC),
            "2010-06-12T00:00:00+00:00",
        ),  # https://code.google.com/p/pyiso8601/issues/detail?id=16
        (
            "1985-04-12T23:20:50.52-05:30",
            datetime.datetime(
                1985, 4, 12, 23, 20, 50, 520000, iso8601.FixedOffset(-5, -30, "-05:30")
            ),
            "1985-04-12T23:20:50.520000-05:30",
        ),  # https://bitbucket.org/micktwomey/pyiso8601/issue/8/015-parses-negative-timezones-incorrectly
        (
            "1997-08-29T06:14:00.000123Z",
            datetime.datetime(1997, 8, 29, 6, 14, 0, 123, iso8601.UTC),
            "1997-08-29T06:14:00.000123+00:00",
        ),  # https://bitbucket.org/micktwomey/pyiso8601/issue/9/regression-parsing-microseconds
        (
            "2014-02",
            datetime.datetime(2014, 2, 1, 0, 0, 0, 0, iso8601.UTC),
            "2014-02-01T00:00:00+00:00",
        ),  # https://bitbucket.org/micktwomey/pyiso8601/issue/14/regression-yyyy-mm-no-longer-parses
        (
            "2014",
            datetime.datetime(2014, 1, 1, 0, 0, 0, 0, iso8601.UTC),
            "2014-01-01T00:00:00+00:00",
        ),  # YYYY
        (
            "1997-08-29T06:14:00,000123Z",
            datetime.datetime(1997, 8, 29, 6, 14, 0, 123, iso8601.UTC),
            "1997-08-29T06:14:00.000123+00:00",
        ),  # Use , as decimal separator
    ],
)
def test_parse_valid_date(
    valid_date: str, expected_datetime: datetime.datetime, isoformat: str
) -> None:
    assert iso8601.is_iso8601(valid_date) is True
    parsed = iso8601.parse_date(valid_date)
    assert parsed.year == expected_datetime.year
    assert parsed.month == expected_datetime.month
    assert parsed.day == expected_datetime.day
    assert parsed.hour == expected_datetime.hour
    assert parsed.minute == expected_datetime.minute
    assert parsed.second == expected_datetime.second
    assert parsed.microsecond == expected_datetime.microsecond
    assert parsed.tzinfo == expected_datetime.tzinfo
    assert parsed == expected_datetime
    assert parsed.isoformat() == expected_datetime.isoformat()
    copy.deepcopy(parsed)  # ensure it's deep copy-able
    pickle.dumps(parsed)  # ensure it pickles
    if isoformat:
        assert parsed.isoformat() == isoformat
    assert iso8601.parse_date(parsed.isoformat()) == parsed  # Test round trip


@hypothesis.given(s=hypothesis.strategies.datetimes())
def test_hypothesis_valid_naive_datetimes(s: datetime.datetime) -> None:
    as_string = s.isoformat()
    parsed = iso8601.parse_date(as_string, default_timezone=None)
    print(f"{s!r} {as_string!r} {parsed!r}")
    assert s == parsed


@hypothesis.given(
    s=hypothesis.strategies.datetimes(timezones=hypothesis.extra.pytz.timezones())
)
def test_hypothesis_valid_datetimes_with_timezone(s: datetime.datetime) -> None:
    as_string = s.isoformat()
    parsed = iso8601.parse_date(as_string)
    print(f"{s!r} {as_string!r} {parsed!r}")
    assert s == parsed



from application.interactor import NCouponInteractor
from custom_sysactions import user_authorized_to_execute_action

from msgbus import TK_SYS_ACK
from mw_helper import show_message_dialog
from sysactions import (
    action,
    StopAction,
)

from typing import List


@action
def remove_applied_benefit_on_payment(
    pos_id,         # type: str
    coupon_ids,     # type: str
):
    # type: (...) -> None

    NCouponInteractor.remove_benefits(
        pos_id=pos_id,
        benefit_keys=[coupon_ids],
    )


@user_authorized_to_execute_action
def remove_discounts(
    pos_id,     # type: str
    coupon_ids  # type: List[str]
):
    # type: (...) -> None

    msg = NCouponInteractor.remove_benefits(
        pos_id=pos_id,
        benefit_keys=coupon_ids,
    )
    if msg.token != TK_SYS_ACK:
        show_message_dialog(
            pos_id=pos_id,
            title="$ERROR",
            message=msg.data,
        )
        raise StopAction()



from xml.etree import cElementTree as eTree

from dailygoals_services.model.repository import DailyGoalsRepository
from dailygoals_services.service import DailyGoalsService
from externals.util import LoggerUtil


logger = LoggerUtil.get_logger_name(__name__)


class UpdateTotalSoldOrderInteractor:
    def __init__(self, daily_goals_repository, daily_goals_service):
        # type: (DailyGoalsRepository, DailyGoalsService) -> None

        self.daily_goals_repository = daily_goals_repository
        self.daily_goals_service = daily_goals_service

    def update(self, data, event_type):
        # type: (str, str) -> None

        if event_type == "PAID":
            logger.info("[handle_event] ORDER_MODIFIED event received")
            order_picture = eTree.XML(data)

            if order_picture.find(".//CustomOrderProperties/OrderProperty[@key='TABLE_ID']") is not None:
                return

            order = order_picture.find('.//Order')
            self.daily_goals_service.update_daily_sold(order)



"""
requests.cookies
~~~~~~~~~~~~~~~~

Compatibility code to be able to use `cookielib.CookieJar` with requests.

requests.utils imports from here, so be careful with imports.
"""

import calendar
import copy
import time

from ._internal_utils import to_native_string
from .compat import Morsel, MutableMapping, cookielib, urlparse, urlunparse

try:
    import threading
except ImportError:
    import dummy_threading as threading


class MockRequest:
    """Wraps a `requests.Request` to mimic a `urllib2.Request`.

    The code in `cookielib.CookieJar` expects this interface in order to correctly
    manage cookie policies, i.e., determine whether a cookie can be set, given the
    domains of the request and the cookie.

    The original request object is read-only. The client is responsible for collecting
    the new headers via `get_new_headers()` and interpreting them appropriately. You
    probably want `get_cookie_header`, defined below.
    """

    def __init__(self, request):
        self._r = request
        self._new_headers = {}
        self.type = urlparse(self._r.url).scheme

    def get_type(self):
        return self.type

    def get_host(self):
        return urlparse(self._r.url).netloc

    def get_origin_req_host(self):
        return self.get_host()

    def get_full_url(self):
        # Only return the response's URL if the user hadn't set the Host
        # header
        if not self._r.headers.get("Host"):
            return self._r.url
        # If they did set it, retrieve it and reconstruct the expected domain
        host = to_native_string(self._r.headers["Host"], encoding="utf-8")
        parsed = urlparse(self._r.url)
        # Reconstruct the URL as we expect it
        return urlunparse(
            [
                parsed.scheme,
                host,
                parsed.path,
                parsed.params,
                parsed.query,
                parsed.fragment,
            ]
        )

    def is_unverifiable(self):
        return True

    def has_header(self, name):
        return name in self._r.headers or name in self._new_headers

    def get_header(self, name, default=None):
        return self._r.headers.get(name, self._new_headers.get(name, default))

    def add_header(self, key, val):
        """cookielib has no legitimate use for this method; add it back if you find one."""
        raise NotImplementedError(
            "Cookie headers should be added with add_unredirected_header()"
        )

    def add_unredirected_header(self, name, value):
        self._new_headers[name] = value

    def get_new_headers(self):
        return self._new_headers

    @property
    def unverifiable(self):
        return self.is_unverifiable()

    @property
    def origin_req_host(self):
        return self.get_origin_req_host()

    @property
    def host(self):
        return self.get_host()


class MockResponse:
    """Wraps a `httplib.HTTPMessage` to mimic a `urllib.addinfourl`.

    ...what? Basically, expose the parsed HTTP headers from the server response
    the way `cookielib` expects to see them.
    """

    def __init__(self, headers):
        """Make a MockResponse for `cookielib` to read.

        :param headers: a httplib.HTTPMessage or analogous carrying the headers
        """
        self._headers = headers

    def info(self):
        return self._headers

    def getheaders(self, name):
        self._headers.getheaders(name)


def extract_cookies_to_jar(jar, request, response):
    """Extract the cookies from the response into a CookieJar.

    :param jar: cookielib.CookieJar (not necessarily a RequestsCookieJar)
    :param request: our own requests.Request object
    :param response: urllib3.HTTPResponse object
    """
    if not (hasattr(response, "_original_response") and response._original_response):
        return
    # the _original_response field is the wrapped httplib.HTTPResponse object,
    req = MockRequest(request)
    # pull out the HTTPMessage with the headers and put it in the mock:
    res = MockResponse(response._original_response.msg)
    jar.extract_cookies(res, req)


def get_cookie_header(jar, request):
    """
    Produce an appropriate Cookie header string to be sent with `request`, or None.

    :rtype: str
    """
    r = MockRequest(request)
    jar.add_cookie_header(r)
    return r.get_new_headers().get("Cookie")


def remove_cookie_by_name(cookiejar, name, domain=None, path=None):
    """Unsets a cookie by name, by default over all domains and paths.

    Wraps CookieJar.clear(), is O(n).
    """
    clearables = []
    for cookie in cookiejar:
        if cookie.name != name:
            continue
        if domain is not None and domain != cookie.domain:
            continue
        if path is not None and path != cookie.path:
            continue
        clearables.append((cookie.domain, cookie.path, cookie.name))

    for domain, path, name in clearables:
        cookiejar.clear(domain, path, name)


class CookieConflictError(RuntimeError):
    """There are two cookies that meet the criteria specified in the cookie jar.
    Use .get and .set and include domain and path args in order to be more specific.
    """


class RequestsCookieJar(cookielib.CookieJar, MutableMapping):
    """Compatibility class; is a cookielib.CookieJar, but exposes a dict
    interface.

    This is the CookieJar we create by default for requests and sessions that
    don't specify one, since some clients may expect response.cookies and
    session.cookies to support dict operations.

    Requests does not use the dict interface internally; it's just for
    compatibility with external client code. All requests code should work
    out of the box with externally provided instances of ``CookieJar``, e.g.
    ``LWPCookieJar`` and ``FileCookieJar``.

    Unlike a regular CookieJar, this class is pickleable.

    .. warning:: dictionary operations that are normally O(1) may be O(n).
    """

    def get(self, name, default=None, domain=None, path=None):
        """Dict-like get() that also supports optional domain and path args in
        order to resolve naming collisions from using one cookie jar over
        multiple domains.

        .. warning:: operation is O(n), not O(1).
        """
        try:
            return self._find_no_duplicates(name, domain, path)
        except KeyError:
            return default

    def set(self, name, value, **kwargs):
        """Dict-like set() that also supports optional domain and path args in
        order to resolve naming collisions from using one cookie jar over
        multiple domains.
        """
        # support client code that unsets cookies by assignment of a None value:
        if value is None:
            remove_cookie_by_name(
                self, name, domain=kwargs.get("domain"), path=kwargs.get("path")
            )
            return

        if isinstance(value, Morsel):
            c = morsel_to_cookie(value)
        else:
            c = create_cookie(name, value, **kwargs)
        self.set_cookie(c)
        return c

    def iterkeys(self):
        """Dict-like iterkeys() that returns an iterator of names of cookies
        from the jar.

        .. seealso:: itervalues() and iteritems().
        """
        for cookie in iter(self):
            yield cookie.name

    def keys(self):
        """Dict-like keys() that returns a list of names of cookies from the
        jar.

        .. seealso:: values() and items().
        """
        return list(self.iterkeys())

    def itervalues(self):
        """Dict-like itervalues() that returns an iterator of values of cookies
        from the jar.

        .. seealso:: iterkeys() and iteritems().
        """
        for cookie in iter(self):
            yield cookie.value

    def values(self):
        """Dict-like values() that returns a list of values of cookies from the
        jar.

        .. seealso:: keys() and items().
        """
        return list(self.itervalues())

    def iteritems(self):
        """Dict-like iteritems() that returns an iterator of name-value tuples
        from the jar.

        .. seealso:: iterkeys() and itervalues().
        """
        for cookie in iter(self):
            yield cookie.name, cookie.value

    def items(self):
        """Dict-like items() that returns a list of name-value tuples from the
        jar. Allows client-code to call ``dict(RequestsCookieJar)`` and get a
        vanilla python dict of key value pairs.

        .. seealso:: keys() and values().
        """
        return list(self.iteritems())

    def list_domains(self):
        """Utility method to list all the domains in the jar."""
        domains = []
        for cookie in iter(self):
            if cookie.domain not in domains:
                domains.append(cookie.domain)
        return domains

    def list_paths(self):
        """Utility method to list all the paths in the jar."""
        paths = []
        for cookie in iter(self):
            if cookie.path not in paths:
                paths.append(cookie.path)
        return paths

    def multiple_domains(self):
        """Returns True if there are multiple domains in the jar.
        Returns False otherwise.

        :rtype: bool
        """
        domains = []
        for cookie in iter(self):
            if cookie.domain is not None and cookie.domain in domains:
                return True
            domains.append(cookie.domain)
        return False  # there is only one domain in jar

    def get_dict(self, domain=None, path=None):
        """Takes as an argument an optional domain and path and returns a plain
        old Python dict of name-value pairs of cookies that meet the
        requirements.

        :rtype: dict
        """
        dictionary = {}
        for cookie in iter(self):
            if (domain is None or cookie.domain == domain) and (
                path is None or cookie.path == path
            ):
                dictionary[cookie.name] = cookie.value
        return dictionary

    def __contains__(self, name):
        try:
            return super().__contains__(name)
        except CookieConflictError:
            return True

    def __getitem__(self, name):
        """Dict-like __getitem__() for compatibility with client code. Throws
        exception if there are more than one cookie with name. In that case,
        use the more explicit get() method instead.

        .. warning:: operation is O(n), not O(1).
        """
        return self._find_no_duplicates(name)

    def __setitem__(self, name, value):
        """Dict-like __setitem__ for compatibility with client code. Throws
        exception if there is already a cookie of that name in the jar. In that
        case, use the more explicit set() method instead.
        """
        self.set(name, value)

    def __delitem__(self, name):
        """Deletes a cookie given a name. Wraps ``cookielib.CookieJar``'s
        ``remove_cookie_by_name()``.
        """
        remove_cookie_by_name(self, name)

    def set_cookie(self, cookie, *args, **kwargs):
        if (
            hasattr(cookie.value, "startswith")
            and cookie.value.startswith('"')
            and cookie.value.endswith('"')
        ):
            cookie.value = cookie.value.replace('\\"', "")
        return super().set_cookie(cookie, *args, **kwargs)

    def update(self, other):
        """Updates this jar with cookies from another CookieJar or dict-like"""
        if isinstance(other, cookielib.CookieJar):
            for cookie in other:
                self.set_cookie(copy.copy(cookie))
        else:
            super().update(other)

    def _find(self, name, domain=None, path=None):
        """Requests uses this method internally to get cookie values.

        If there are conflicting cookies, _find arbitrarily chooses one.
        See _find_no_duplicates if you want an exception thrown if there are
        conflicting cookies.

        :param name: a string containing name of cookie
        :param domain: (optional) string containing domain of cookie
        :param path: (optional) string containing path of cookie
        :return: cookie.value
        """
        for cookie in iter(self):
            if cookie.name == name:
                if domain is None or cookie.domain == domain:
                    if path is None or cookie.path == path:
                        return cookie.value

        raise KeyError(f"name={name!r}, domain={domain!r}, path={path!r}")

    def _find_no_duplicates(self, name, domain=None, path=None):
        """Both ``__get_item__`` and ``get`` call this function: it's never
        used elsewhere in Requests.

        :param name: a string containing name of cookie
        :param domain: (optional) string containing domain of cookie
        :param path: (optional) string containing path of cookie
        :raises KeyError: if cookie is not found
        :raises CookieConflictError: if there are multiple cookies
            that match name and optionally domain and path
        :return: cookie.value
        """
        toReturn = None
        for cookie in iter(self):
            if cookie.name == name:
                if domain is None or cookie.domain == domain:
                    if path is None or cookie.path == path:
                        if toReturn is not None:
                            # if there are multiple cookies that meet passed in criteria
                            raise CookieConflictError(
                                f"There are multiple cookies with name, {name!r}"
                            )
                        # we will eventually return this as long as no cookie conflict
                        toReturn = cookie.value

        if toReturn:
            return toReturn
        raise KeyError(f"name={name!r}, domain={domain!r}, path={path!r}")

    def __getstate__(self):
        """Unlike a normal CookieJar, this class is pickleable."""
        state = self.__dict__.copy()
        # remove the unpickleable RLock object
        state.pop("_cookies_lock")
        return state

    def __setstate__(self, state):
        """Unlike a normal CookieJar, this class is pickleable."""
        self.__dict__.update(state)
        if "_cookies_lock" not in self.__dict__:
            self._cookies_lock = threading.RLock()

    def copy(self):
        """Return a copy of this RequestsCookieJar."""
        new_cj = RequestsCookieJar()
        new_cj.set_policy(self.get_policy())
        new_cj.update(self)
        return new_cj

    def get_policy(self):
        """Return the CookiePolicy instance used."""
        return self._policy


def _copy_cookie_jar(jar):
    if jar is None:
        return None

    if hasattr(jar, "copy"):
        # We're dealing with an instance of RequestsCookieJar
        return jar.copy()
    # We're dealing with a generic CookieJar instance
    new_jar = copy.copy(jar)
    new_jar.clear()
    for cookie in jar:
        new_jar.set_cookie(copy.copy(cookie))
    return new_jar


def create_cookie(name, value, **kwargs):
    """Make a cookie from underspecified parameters.

    By default, the pair of `name` and `value` will be set for the domain ''
    and sent on every request (this is sometimes called a "supercookie").
    """
    result = {
        "version": 0,
        "name": name,
        "value": value,
        "port": None,
        "domain": "",
        "path": "/",
        "secure": False,
        "expires": None,
        "discard": True,
        "comment": None,
        "comment_url": None,
        "rest": {"HttpOnly": None},
        "rfc2109": False,
    }

    badargs = set(kwargs) - set(result)
    if badargs:
        raise TypeError(
            f"create_cookie() got unexpected keyword arguments: {list(badargs)}"
        )

    result.update(kwargs)
    result["port_specified"] = bool(result["port"])
    result["domain_specified"] = bool(result["domain"])
    result["domain_initial_dot"] = result["domain"].startswith(".")
    result["path_specified"] = bool(result["path"])

    return cookielib.Cookie(**result)


def morsel_to_cookie(morsel):
    """Convert a Morsel object into a Cookie containing the one k/v pair."""

    expires = None
    if morsel["max-age"]:
        try:
            expires = int(time.time() + int(morsel["max-age"]))
        except ValueError:
            raise TypeError(f"max-age: {morsel['max-age']} must be integer")
    elif morsel["expires"]:
        time_template = "%a, %d-%b-%Y %H:%M:%S GMT"
        expires = calendar.timegm(time.strptime(morsel["expires"], time_template))
    return create_cookie(
        comment=morsel["comment"],
        comment_url=bool(morsel["comment"]),
        discard=False,
        domain=morsel["domain"],
        expires=expires,
        name=morsel.key,
        path=morsel["path"],
        port=None,
        rest={"HttpOnly": morsel["httponly"]},
        rfc2109=False,
        secure=bool(morsel["secure"]),
        value=morsel.value,
        version=morsel["version"] or 0,
    )


def cookiejar_from_dict(cookie_dict, cookiejar=None, overwrite=True):
    """Returns a CookieJar from a key/value dictionary.

    :param cookie_dict: Dict of key/values to insert into CookieJar.
    :param cookiejar: (optional) A cookiejar to add the cookies to.
    :param overwrite: (optional) If False, will not replace cookies
        already in the jar with new ones.
    :rtype: CookieJar
    """
    if cookiejar is None:
        cookiejar = RequestsCookieJar()

    if cookie_dict is not None:
        names_from_jar = [cookie.name for cookie in cookiejar]
        for name in cookie_dict:
            if overwrite or (name not in names_from_jar):
                cookiejar.set_cookie(create_cookie(name, cookie_dict[name]))

    return cookiejar


def merge_cookies(cookiejar, cookies):
    """Add cookies to cookiejar and returns a merged CookieJar.

    :param cookiejar: CookieJar object to add the cookies to.
    :param cookies: Dictionary or CookieJar object to be added.
    :rtype: CookieJar
    """
    if not isinstance(cookiejar, cookielib.CookieJar):
        raise ValueError("You can only merge into CookieJar")

    if isinstance(cookies, dict):
        cookiejar = cookiejar_from_dict(cookies, cookiejar=cookiejar, overwrite=False)
    elif isinstance(cookies, cookielib.CookieJar):
        try:
            cookiejar.update(cookies)
        except AttributeError:
            for cookie_in_jar in cookies:
                cookiejar.set_cookie(cookie_in_jar)

    return cookiejar



from logging import Logger

from typing import Dict

from production.box import ItemCustomParamsFilterBox
from production.treebuilder import BoxConfiguration
from production.repository import ProductRepository
from ._BoxBuilder import BoxBuilder


class ItemCustomParamsFilterBoxBuilder(BoxBuilder):
    def __init__(
        self,                   # type: ItemCustomParamsFilterBoxBuilder
        product_repository,     # type: ProductRepository
        loggers,                # type: Dict[str, Logger]
    ):
        # type: (...) -> None

        super(ItemCustomParamsFilterBoxBuilder, self).__init__(loggers=loggers)

        self.product_repository = product_repository

    def build(
        self,       # type: ItemCustomParamsFilterBoxBuilder
        config,     # type: BoxConfiguration
    ):
        # type: (...) -> ItemCustomParamsFilterBox

        return ItemCustomParamsFilterBox(
            name=config.name,
            sons=config.sons,
            product_repository=self.product_repository,
            custom_params=self.get_extra(config, "CustomParams", None),
            inherit_comments=self.get_extra(config, "InheritComments", "false").lower() == "true",
            logger=self.get_logger(config=config),
        )



from ._DefaultManagerConfig import (
    DefaultManagerConfig,
    Window,
)
from ._CatalogConfigurations import CatalogConfigurations
from ._MediaConfigurations import MediaConfigurations
from ._UserConfigurations import UserConfigurations
from ._Configurations import Configurations
from ._LoaderConfigurations import LoaderConfigurations
from ._BinaryConfigurations import BinaryConfigurations
from ._SystemPaths import SystemPaths



# bn.py - functions for handling Canadian Business Numbers (BNs)
#
# Copyright (C) 2017 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""BN (Canadian Business Number).

A Business Number (BN) is a 9-digit identification number for businesses
issued by the Canada Revenue Agency for tax purposes. The 9-digit number can
be followed by two letters (program identifier) and 4 digits (reference
number) to form a program account (or BN15).

More information:

* https://www.canada.ca/en/services/taxes/business-number.html
* https://www.ic.gc.ca/app/scr/cc/CorporationsCanada/fdrlCrpSrch.html?locale=en_CA/

>>> validate('12302 6635')
'123026635'
>>> validate('12302 6635 RC 0001')
'123026635RC0001'
>>> validate('123456783')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('12345678Z')
Traceback (most recent call last):
    ...
InvalidFormat: ...
"""

from stdnum import luhn
from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, '- ').strip()


def validate(number):
    """Check if the number is a valid BN or BN15. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if len(number) not in (9, 15):
        raise InvalidLength()
    if not isdigits(number[:9]):
        raise InvalidFormat()
    luhn.validate(number[:9])
    if len(number) == 15:
        if number[9:11] not in ('RC', 'RM', 'RP', 'RT'):
            raise InvalidComponent()
        if not isdigits(number[11:]):
            raise InvalidFormat()
    return number


def is_valid(number):
    """Check if the number is a valid BN or BN15."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



from __future__ import annotations

import collections
from collections.abc import Iterable
import textwrap
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import overload
from typing import Sequence
from typing import Tuple
from typing import TypeVar
from typing import Union
import uuid
import warnings

from sqlalchemy.util import asbool  # noqa
from sqlalchemy.util import immutabledict  # noqa
from sqlalchemy.util import memoized_property  # noqa
from sqlalchemy.util import to_list  # noqa
from sqlalchemy.util import unique_list  # noqa

from .compat import inspect_getfullargspec


EMPTY_DICT: Mapping[Any, Any] = immutabledict()
_T = TypeVar("_T")


class _ModuleClsMeta(type):
    def __setattr__(cls, key: str, value: Callable) -> None:
        super().__setattr__(key, value)
        cls._update_module_proxies(key)  # type: ignore


class ModuleClsProxy(metaclass=_ModuleClsMeta):
    """Create module level proxy functions for the
    methods on a given class.

    The functions will have a compatible signature
    as the methods.

    """

    _setups: Dict[type, Tuple[set, list]] = collections.defaultdict(
        lambda: (set(), [])
    )

    @classmethod
    def _update_module_proxies(cls, name: str) -> None:
        attr_names, modules = cls._setups[cls]
        for globals_, locals_ in modules:
            cls._add_proxied_attribute(name, globals_, locals_, attr_names)

    def _install_proxy(self) -> None:
        attr_names, modules = self._setups[self.__class__]
        for globals_, locals_ in modules:
            globals_["_proxy"] = self
            for attr_name in attr_names:
                globals_[attr_name] = getattr(self, attr_name)

    def _remove_proxy(self) -> None:
        attr_names, modules = self._setups[self.__class__]
        for globals_, locals_ in modules:
            globals_["_proxy"] = None
            for attr_name in attr_names:
                del globals_[attr_name]

    @classmethod
    def create_module_class_proxy(cls, globals_, locals_):
        attr_names, modules = cls._setups[cls]
        modules.append((globals_, locals_))
        cls._setup_proxy(globals_, locals_, attr_names)

    @classmethod
    def _setup_proxy(cls, globals_, locals_, attr_names):
        for methname in dir(cls):
            cls._add_proxied_attribute(methname, globals_, locals_, attr_names)

    @classmethod
    def _add_proxied_attribute(cls, methname, globals_, locals_, attr_names):
        if not methname.startswith("_"):
            meth = getattr(cls, methname)
            if callable(meth):
                locals_[methname] = cls._create_method_proxy(
                    methname, globals_, locals_
                )
            else:
                attr_names.add(methname)

    @classmethod
    def _create_method_proxy(cls, name, globals_, locals_):
        fn = getattr(cls, name)

        def _name_error(name, from_):
            raise NameError(
                "Can't invoke function '%s', as the proxy object has "
                "not yet been "
                "established for the Alembic '%s' class.  "
                "Try placing this code inside a callable."
                % (name, cls.__name__)
            ) from from_

        globals_["_name_error"] = _name_error

        translations = getattr(fn, "_legacy_translations", [])
        if translations:
            spec = inspect_getfullargspec(fn)
            if spec[0] and spec[0][0] == "self":
                spec[0].pop(0)

            outer_args = inner_args = "*args, **kw"
            translate_str = "args, kw = _translate(%r, %r, %r, args, kw)" % (
                fn.__name__,
                tuple(spec),
                translations,
            )

            def translate(fn_name, spec, translations, args, kw):
                return_kw = {}
                return_args = []

                for oldname, newname in translations:
                    if oldname in kw:
                        warnings.warn(
                            "Argument %r is now named %r "
                            "for method %s()." % (oldname, newname, fn_name)
                        )
                        return_kw[newname] = kw.pop(oldname)
                return_kw.update(kw)

                args = list(args)
                if spec[3]:
                    pos_only = spec[0][: -len(spec[3])]
                else:
                    pos_only = spec[0]
                for arg in pos_only:
                    if arg not in return_kw:
                        try:
                            return_args.append(args.pop(0))
                        except IndexError:
                            raise TypeError(
                                "missing required positional argument: %s"
                                % arg
                            )
                return_args.extend(args)

                return return_args, return_kw

            globals_["_translate"] = translate
        else:
            outer_args = "*args, **kw"
            inner_args = "*args, **kw"
            translate_str = ""

        func_text = textwrap.dedent(
            """\
        def %(name)s(%(args)s):
            %(doc)r
            %(translate)s
            try:
                p = _proxy
            except NameError as ne:
                _name_error('%(name)s', ne)
            return _proxy.%(name)s(%(apply_kw)s)
            e
        """
            % {
                "name": name,
                "translate": translate_str,
                "args": outer_args,
                "apply_kw": inner_args,
                "doc": fn.__doc__,
            }
        )
        lcl = {}

        exec(func_text, globals_, lcl)
        return lcl[name]


def _with_legacy_names(translations):
    def decorate(fn):
        fn._legacy_translations = translations
        return fn

    return decorate


def rev_id() -> str:
    return uuid.uuid4().hex[-12:]


@overload
def to_tuple(x: Any, default: tuple) -> tuple:
    ...


@overload
def to_tuple(x: None, default: Optional[_T] = None) -> _T:
    ...


@overload
def to_tuple(x: Any, default: Optional[tuple] = None) -> tuple:
    ...


def to_tuple(x, default=None):
    if x is None:
        return default
    elif isinstance(x, str):
        return (x,)
    elif isinstance(x, Iterable):
        return tuple(x)
    else:
        return (x,)


def dedupe_tuple(tup: Tuple[str, ...]) -> Tuple[str, ...]:
    return tuple(unique_list(tup))


class Dispatcher:
    def __init__(self, uselist: bool = False) -> None:
        self._registry: Dict[tuple, Any] = {}
        self.uselist = uselist

    def dispatch_for(
        self, target: Any, qualifier: str = "default"
    ) -> Callable:
        def decorate(fn):
            if self.uselist:
                self._registry.setdefault((target, qualifier), []).append(fn)
            else:
                assert (target, qualifier) not in self._registry
                self._registry[(target, qualifier)] = fn
            return fn

        return decorate

    def dispatch(self, obj: Any, qualifier: str = "default") -> Any:
        if isinstance(obj, str):
            targets: Sequence = [obj]
        elif isinstance(obj, type):
            targets = obj.__mro__
        else:
            targets = type(obj).__mro__

        for spcls in targets:
            if qualifier != "default" and (spcls, qualifier) in self._registry:
                return self._fn_or_list(self._registry[(spcls, qualifier)])
            elif (spcls, "default") in self._registry:
                return self._fn_or_list(self._registry[(spcls, "default")])
        else:
            raise ValueError("no dispatch function for object: %s" % obj)

    def _fn_or_list(
        self, fn_or_list: Union[List[Callable], Callable]
    ) -> Callable:
        if self.uselist:

            def go(*arg, **kw):
                for fn in fn_or_list:
                    fn(*arg, **kw)

            return go
        else:
            return fn_or_list  # type: ignore

    def branch(self) -> Dispatcher:
        """Return a copy of this dispatcher that is independently
        writable."""

        d = Dispatcher()
        if self.uselist:
            d._registry.update(
                (k, [fn for fn in self._registry[k]]) for k in self._registry
            )
        else:
            d._registry.update(self._registry)
        return d


def not_none(value: Optional[_T]) -> _T:
    assert value is not None
    return value



"""distutils.bcppcompiler

Contains BorlandCCompiler, an implementation of the abstract CCompiler class
for the Borland C++ compiler.
"""

# This implementation by Lyle Johnson, based on the original msvccompiler.py
# module and using the directions originally published by Gordon Williams.

# XXX looks like there's a LOT of overlap between these two classes:
# someone should sit down and factor out the common code as
# WindowsCCompiler!  --GPW

import os
import warnings

from ._log import log
from ._modified import newer
from .ccompiler import CCompiler, gen_preprocess_options
from .errors import (
    CompileError,
    DistutilsExecError,
    LibError,
    LinkError,
    UnknownFileError,
)
from .file_util import write_file

warnings.warn(
    "bcppcompiler is deprecated and slated to be removed "
    "in the future. Please discontinue use or file an issue "
    "with pypa/distutils describing your use case.",
    DeprecationWarning,
)


class BCPPCompiler(CCompiler):
    """Concrete class that implements an interface to the Borland C/C++
    compiler, as defined by the CCompiler abstract class.
    """

    compiler_type = 'bcpp'

    # Just set this so CCompiler's constructor doesn't barf.  We currently
    # don't use the 'set_executables()' bureaucracy provided by CCompiler,
    # as it really isn't necessary for this sort of single-compiler class.
    # Would be nice to have a consistent interface with UnixCCompiler,
    # though, so it's worth thinking about.
    executables = {}

    # Private class data (need to distinguish C from C++ source for compiler)
    _c_extensions = ['.c']
    _cpp_extensions = ['.cc', '.cpp', '.cxx']

    # Needed for the filename generation methods provided by the
    # base class, CCompiler.
    src_extensions = _c_extensions + _cpp_extensions
    obj_extension = '.obj'
    static_lib_extension = '.lib'
    shared_lib_extension = '.dll'
    static_lib_format = shared_lib_format = '%s%s'
    exe_extension = '.exe'

    def __init__(self, verbose=False, dry_run=False, force=False):
        super().__init__(verbose, dry_run, force)

        # These executables are assumed to all be in the path.
        # Borland doesn't seem to use any special registry settings to
        # indicate their installation locations.

        self.cc = "bcc32.exe"
        self.linker = "ilink32.exe"
        self.lib = "tlib.exe"

        self.preprocess_options = None
        self.compile_options = ['/tWM', '/O2', '/q', '/g0']
        self.compile_options_debug = ['/tWM', '/Od', '/q', '/g0']

        self.ldflags_shared = ['/Tpd', '/Gn', '/q', '/x']
        self.ldflags_shared_debug = ['/Tpd', '/Gn', '/q', '/x']
        self.ldflags_static = []
        self.ldflags_exe = ['/Gn', '/q', '/x']
        self.ldflags_exe_debug = ['/Gn', '/q', '/x', '/r']

    # -- Worker methods ------------------------------------------------

    def compile(
        self,
        sources,
        output_dir=None,
        macros=None,
        include_dirs=None,
        debug=False,
        extra_preargs=None,
        extra_postargs=None,
        depends=None,
    ):
        macros, objects, extra_postargs, pp_opts, build = self._setup_compile(
            output_dir, macros, include_dirs, sources, depends, extra_postargs
        )
        compile_opts = extra_preargs or []
        compile_opts.append('-c')
        if debug:
            compile_opts.extend(self.compile_options_debug)
        else:
            compile_opts.extend(self.compile_options)

        for obj in objects:
            try:
                src, ext = build[obj]
            except KeyError:
                continue
            # XXX why do the normpath here?
            src = os.path.normpath(src)
            obj = os.path.normpath(obj)
            # XXX _setup_compile() did a mkpath() too but before the normpath.
            # Is it possible to skip the normpath?
            self.mkpath(os.path.dirname(obj))

            if ext == '.res':
                # This is already a binary file -- skip it.
                continue  # the 'for' loop
            if ext == '.rc':
                # This needs to be compiled to a .res file -- do it now.
                try:
                    self.spawn(["brcc32", "-fo", obj, src])
                except DistutilsExecError as msg:
                    raise CompileError(msg)
                continue  # the 'for' loop

            # The next two are both for the real compiler.
            if ext in self._c_extensions:
                input_opt = ""
            elif ext in self._cpp_extensions:
                input_opt = "-P"
            else:
                # Unknown file type -- no extra options.  The compiler
                # will probably fail, but let it just in case this is a
                # file the compiler recognizes even if we don't.
                input_opt = ""

            output_opt = "-o" + obj

            # Compiler command line syntax is: "bcc32 [options] file(s)".
            # Note that the source file names must appear at the end of
            # the command line.
            try:
                self.spawn(
                    [self.cc]
                    + compile_opts
                    + pp_opts
                    + [input_opt, output_opt]
                    + extra_postargs
                    + [src]
                )
            except DistutilsExecError as msg:
                raise CompileError(msg)

        return objects

    # compile ()

    def create_static_lib(
        self, objects, output_libname, output_dir=None, debug=False, target_lang=None
    ):
        (objects, output_dir) = self._fix_object_args(objects, output_dir)
        output_filename = self.library_filename(output_libname, output_dir=output_dir)

        if self._need_link(objects, output_filename):
            lib_args = [output_filename, '/u'] + objects
            if debug:
                pass  # XXX what goes here?
            try:
                self.spawn([self.lib] + lib_args)
            except DistutilsExecError as msg:
                raise LibError(msg)
        else:
            log.debug("skipping %s (up-to-date)", output_filename)

    # create_static_lib ()

    def link(  # noqa: C901
        self,
        target_desc,
        objects,
        output_filename,
        output_dir=None,
        libraries=None,
        library_dirs=None,
        runtime_library_dirs=None,
        export_symbols=None,
        debug=False,
        extra_preargs=None,
        extra_postargs=None,
        build_temp=None,
        target_lang=None,
    ):
        # XXX this ignores 'build_temp'!  should follow the lead of
        # msvccompiler.py

        (objects, output_dir) = self._fix_object_args(objects, output_dir)
        (libraries, library_dirs, runtime_library_dirs) = self._fix_lib_args(
            libraries, library_dirs, runtime_library_dirs
        )

        if runtime_library_dirs:
            log.warning(
                "I don't know what to do with 'runtime_library_dirs': %s",
                str(runtime_library_dirs),
            )

        if output_dir is not None:
            output_filename = os.path.join(output_dir, output_filename)

        if self._need_link(objects, output_filename):
            # Figure out linker args based on type of target.
            if target_desc == CCompiler.EXECUTABLE:
                startup_obj = 'c0w32'
                if debug:
                    ld_args = self.ldflags_exe_debug[:]
                else:
                    ld_args = self.ldflags_exe[:]
            else:
                startup_obj = 'c0d32'
                if debug:
                    ld_args = self.ldflags_shared_debug[:]
                else:
                    ld_args = self.ldflags_shared[:]

            # Create a temporary exports file for use by the linker
            if export_symbols is None:
                def_file = ''
            else:
                head, tail = os.path.split(output_filename)
                modname, ext = os.path.splitext(tail)
                temp_dir = os.path.dirname(objects[0])  # preserve tree structure
                def_file = os.path.join(temp_dir, f'{modname}.def')
                contents = ['EXPORTS']
                for sym in export_symbols or []:
                    contents.append(f'  {sym}=_{sym}')
                self.execute(write_file, (def_file, contents), f"writing {def_file}")

            # Borland C++ has problems with '/' in paths
            objects2 = map(os.path.normpath, objects)
            # split objects in .obj and .res files
            # Borland C++ needs them at different positions in the command line
            objects = [startup_obj]
            resources = []
            for file in objects2:
                (base, ext) = os.path.splitext(os.path.normcase(file))
                if ext == '.res':
                    resources.append(file)
                else:
                    objects.append(file)

            for ell in library_dirs:
                ld_args.append(f"/L{os.path.normpath(ell)}")
            ld_args.append("/L.")  # we sometimes use relative paths

            # list of object files
            ld_args.extend(objects)

            # XXX the command-line syntax for Borland C++ is a bit wonky;
            # certain filenames are jammed together in one big string, but
            # comma-delimited.  This doesn't mesh too well with the
            # Unix-centric attitude (with a DOS/Windows quoting hack) of
            # 'spawn()', so constructing the argument list is a bit
            # awkward.  Note that doing the obvious thing and jamming all
            # the filenames and commas into one argument would be wrong,
            # because 'spawn()' would quote any filenames with spaces in
            # them.  Arghghh!.  Apparently it works fine as coded...

            # name of dll/exe file
            ld_args.extend([',', output_filename])
            # no map file and start libraries
            ld_args.append(',,')

            for lib in libraries:
                # see if we find it and if there is a bcpp specific lib
                # (xxx_bcpp.lib)
                libfile = self.find_library_file(library_dirs, lib, debug)
                if libfile is None:
                    ld_args.append(lib)
                    # probably a BCPP internal library -- don't warn
                else:
                    # full name which prefers bcpp_xxx.lib over xxx.lib
                    ld_args.append(libfile)

            # some default libraries
            ld_args.extend(('import32', 'cw32mt'))

            # def file for export symbols
            ld_args.extend([',', def_file])
            # add resource files
            ld_args.append(',')
            ld_args.extend(resources)

            if extra_preargs:
                ld_args[:0] = extra_preargs
            if extra_postargs:
                ld_args.extend(extra_postargs)

            self.mkpath(os.path.dirname(output_filename))
            try:
                self.spawn([self.linker] + ld_args)
            except DistutilsExecError as msg:
                raise LinkError(msg)

        else:
            log.debug("skipping %s (up-to-date)", output_filename)

    # link ()

    # -- Miscellaneous methods -----------------------------------------

    def find_library_file(self, dirs, lib, debug=False):
        # List of effective library names to try, in order of preference:
        # xxx_bcpp.lib is better than xxx.lib
        # and xxx_d.lib is better than xxx.lib if debug is set
        #
        # The "_bcpp" suffix is to handle a Python installation for people
        # with multiple compilers (primarily Distutils hackers, I suspect
        # ;-).  The idea is they'd have one static library for each
        # compiler they care about, since (almost?) every Windows compiler
        # seems to have a different format for static libraries.
        if debug:
            dlib = lib + "_d"
            try_names = (dlib + "_bcpp", lib + "_bcpp", dlib, lib)
        else:
            try_names = (lib + "_bcpp", lib)

        for dir in dirs:
            for name in try_names:
                libfile = os.path.join(dir, self.library_filename(name))
                if os.path.exists(libfile):
                    return libfile
        else:
            # Oops, didn't find it in *any* of 'dirs'
            return None

    # overwrite the one from CCompiler to support rc and res-files
    def object_filenames(self, source_filenames, strip_dir=False, output_dir=''):
        if output_dir is None:
            output_dir = ''
        obj_names = []
        for src_name in source_filenames:
            # use normcase to make sure '.rc' is really '.rc' and not '.RC'
            (base, ext) = os.path.splitext(os.path.normcase(src_name))
            if ext not in (self.src_extensions + ['.rc', '.res']):
                raise UnknownFileError(f"unknown file type '{ext}' (from '{src_name}')")
            if strip_dir:
                base = os.path.basename(base)
            if ext == '.res':
                # these can go unchanged
                obj_names.append(os.path.join(output_dir, base + ext))
            elif ext == '.rc':
                # these need to be compiled to .res-files
                obj_names.append(os.path.join(output_dir, base + '.res'))
            else:
                obj_names.append(os.path.join(output_dir, base + self.obj_extension))
        return obj_names

    # object_filenames ()

    def preprocess(
        self,
        source,
        output_file=None,
        macros=None,
        include_dirs=None,
        extra_preargs=None,
        extra_postargs=None,
    ):
        (_, macros, include_dirs) = self._fix_compile_args(None, macros, include_dirs)
        pp_opts = gen_preprocess_options(macros, include_dirs)
        pp_args = ['cpp32.exe'] + pp_opts
        if output_file is not None:
            pp_args.append('-o' + output_file)
        if extra_preargs:
            pp_args[:0] = extra_preargs
        if extra_postargs:
            pp_args.extend(extra_postargs)
        pp_args.append(source)

        # We need to preprocess: either we're being forced to, or the
        # source file is newer than the target (or the target doesn't
        # exist).
        if self.force or output_file is None or newer(source, output_file):
            if output_file:
                self.mkpath(os.path.dirname(output_file))
            try:
                self.spawn(pp_args)
            except DistutilsExecError as msg:
                print(msg)
                raise CompileError(msg)

    # preprocess()



class OrderWaitingProcessingException(Exception):

    def __init__(
        self,       # type: OrderWaitingProcessingException
    ):
        # type: (...) -> None

        self.message = "Pedido na fila para processamento"



import base64
import json
from xml.etree import cElementTree as eTree

from typing import Dict, Any, List

from nreports_api.shared import JSON
from .._DataBuilder import DataBuilder


class KitchenLabelBuilder(DataBuilder):

    def build_data(
        self,
        data_to_build: Dict[str, JSON],
        **kwargs: Any,
    ) -> Dict[str, JSON]:

        production_order = eTree.XML(base64.b64decode(kwargs["productionOrder"] + '===='))
        order_id = production_order.get("order_id")
        items = production_order.findall(".//Item")

        data_by_labels = []
        for item in items:
            item_data_labels = self._get_labels_from_item(item)
            data_by_labels.extend(item_data_labels)

        total_labels = len(data_by_labels)
        for index, label in enumerate(data_by_labels):
            label["CURRENT_LABEL"] = index + 1
            label["TOTAL_LABELS"] = total_labels
            label["ORDER_ID"] = order_id

        data_to_build["labels"] = data_by_labels
        return data_to_build

    @staticmethod
    def _get_labels_from_item(
        item: eTree.Element
    ) -> List[JSON]:

        label_tag = item.find("./CustomParams/CustomParam[@id='KitchenLabels']")
        if label_tag is None:
            return []

        item_comment = item.find("Comment").get("comment") if item.find("Comment") is not None else ""
        item_labels = json.loads(label_tag.get("value"))

        data_by_labels = []
        initial_index = len(data_by_labels)
        for index, label in enumerate(item_labels, start=initial_index):
            item_comment = item_comment if label.get("observation", False) is True else ""
            label_data = {"LABEL_TEXT": label.get("text", ""), "OBSERVATION": item_comment}

            data_by_labels.append(label_data)

        return data_by_labels



from datetime import datetime
from logging import Logger

from production.model import Item
from typing import (
    Optional,
    List,
)

from ._OrderChangerProductionBox import OrderChangerProductionBox
from ..model import (
    ProductionOrder,
    StateEvent,
    create_states_from_state,
)

timestamp_format = "%Y-%m-%dT%H:%M:%S.%f"


class FilterDeletedItemsNotSentToKitchenBox(OrderChangerProductionBox):

    def __init__(
        self,                            # type: FilterDeletedItemsNotSentToKitchenBox
        name,                            # type: str
        sons,                            # type: object
        state_to_delete_voided_items,    # type: str
        logger=None,                     # type: Optional[Logger]
    ):
        super(FilterDeletedItemsNotSentToKitchenBox, self).__init__(name, sons, logger)

        self.state_to_delete_voided_items = create_states_from_state(state=state_to_delete_voided_items)

    def change_order(
        self,     # type: FilterDeletedItemsNotSentToKitchenBox
        order,    # type: ProductionOrder
    ):
        # type: (...) -> ProductionOrder

        self._remove_deleted_items_in_last_in_progress_state(order=order)
        return order

    def _remove_deleted_items_in_last_in_progress_state(
        self,     # type: FilterDeletedItemsNotSentToKitchenBox
        order,    # type: ProductionOrder
    ):
        # type: (...) -> None

        filtered_states = self._get_filtered_states(order=order)
        last_state_datetime = datetime.strptime(
            order.created_at,
            timestamp_format,
        )
        for i, state in enumerate(filtered_states):
            state_datetime = datetime.strptime(
                state.timestamp_gmt,
                timestamp_format,
            )
            self._remove_empty_items_before_last_state(
                items=order.items,
                last_state_datetime=last_state_datetime,
                current_state_datetime=state_datetime,
            )
            last_state_datetime = state_datetime

    def _remove_empty_items_before_last_state(
        self,                       # type: FilterDeletedItemsNotSentToKitchenBox
        items,                      # type: List[Item]
        last_state_datetime,        # type: datetime
        current_state_datetime,     # type: datetime
    ):
        # type: (...) -> None

        for item in items[:]:
            if item.qty == 0:
                line_created_at = datetime.strptime(
                    item.created_at,
                    timestamp_format,
                )

                if line_created_at >= last_state_datetime:
                    line_last_updated = datetime.strptime(
                        item.last_updated,
                        timestamp_format + "Z",
                    )

                    if line_last_updated <= current_state_datetime:
                        items.remove(item)

            self._remove_empty_items_before_last_state(
                items=item.items,
                last_state_datetime=last_state_datetime,
                current_state_datetime=current_state_datetime,
            )

    @staticmethod
    def _get_next_state_datetime(
        current_index,    # type: int
        states,           # type: List[StateEvent]
    ):
        # type: (...) -> Optional[datetime]

        if current_index + 1 < len(states):
            next_state = states[current_index + 1]
            return datetime.strptime(
                next_state.timestamp_gmt,
                timestamp_format,
            )

        return None

    def _get_filtered_states(
        self,   # type: FilterDeletedItemsNotSentToKitchenBox
        order,  # type: ProductionOrder
    ):
        # type: (...) -> List[StateEvent]

        filtered_states = []
        for state in order.state_history:
            if state.state in self.state_to_delete_voided_items:
                filtered_states.append(state)

        return sorted(
            filtered_states,
            key=lambda item: item.timestamp_gmt,
        )



# Copyright 2009-2017 Wander Lairson Costa
# Copyright 2009-2020 PyUSB contributors
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

__author__ = 'Wander Lairson Costa'

__all__ = ['methodtrace', 'functiontrace']

import logging
import usb._interop as _interop

_enable_tracing = False

def enable_tracing(enable):
    global _enable_tracing
    _enable_tracing = enable

def _trace_function_call(logger, fname, *args, **named_args):
    logger.debug(
                # TODO: check if 'f' is a method or a free function
                fname + '(' + \
                ', '.join((str(val) for val in args)) + \
                ', '.join((name + '=' + str(val) for name, val in named_args.items())) + ')'
            )

# decorator for methods calls tracing
def methodtrace(logger):
    def decorator_logging(f):
        if not _enable_tracing:
            return f
        def do_trace(*args, **named_args):
            # this if is just a optimization to avoid unecessary string formatting
            if logging.DEBUG >= logger.getEffectiveLevel():
                fn = type(args[0]).__name__ + '.' + f.__name__
                _trace_function_call(logger, fn, *args[1:], **named_args)
            return f(*args, **named_args)
        _interop._update_wrapper(do_trace, f)
        return do_trace
    return decorator_logging

# decorator for methods calls tracing
def functiontrace(logger):
    def decorator_logging(f):
        if not _enable_tracing:
            return f
        def do_trace(*args, **named_args):
            # this if is just a optimization to avoid unecessary string formatting
            if logging.DEBUG >= logger.getEffectiveLevel():
                _trace_function_call(logger, f.__name__, *args, **named_args)
            return f(*args, **named_args)
        _interop._update_wrapper(do_trace, f)
        return do_trace
    return decorator_logging



import json
import logging
import os
from typing import (
    Dict,
    List,
    Optional,
    Union,
)

from cfgtools import read
from filesystem import FileSystem
from message_handler_builder import Configs
from nreports.utils import pickle_deep_copy
from nreports_api.shared import JSON
from systools import sys_parsefilepath
from ._IncludeCycle import IncludeCycle

logger = logging.getLogger(name=__name__)


class Configurator(object):

    def __init__(
        self,
        loader_path: str,
        file_system: FileSystem,
    ) -> None:

        self.loader_path = loader_path
        self.file_system = file_system

        self.reader = read(file_or_path=loader_path)
        self.group = Configs.component_name + "."
        self._default_images_path = os.path.join("..", "src", "nreports", "lib", "images")

    def get_configs(
        self,
    ) -> Configs:

        database_dir = os.path.join(os.path.dirname(self.loader_path), "..", "..", "data", "server", "databases")
        report_files = self._read_report_files()
        extra_modules = self.reader.find_values(path=self.group + "BuilderExtraModules") or []
        images_path = self._read_images_path()
        constants = self._read_constants()

        return Configs(
            database_dir=database_dir,
            report_files=report_files,
            images_path=images_path,
            constants=constants,
            builder_extra_modules=extra_modules,
        )

    def get_required_services(
        self,
    ) -> str:
        return "|".join(self.reader.find_values(path=self.group + "RequiredServices") or [])

    def _find(
        self,
        key: str,
        default_value: Optional[str] = None,
    ) -> str:

        return self.reader.find_value(
            path=self.group + key,
            default=default_value,
        )

    def _read_report_files(
        self,
    ) -> Dict[str, List[JSON]]:

        report_files_dirs = [os.path.join("..", "src", "nreports", "reports")]
        extra_report_files_dirs = self.reader.find_values(path=self.group + "ExtraReportFileDir")
        if extra_report_files_dirs is None:
            extra_report_files_dirs = []
        report_files_dirs.extend(extra_report_files_dirs)

        report_files_dirs = map(sys_parsefilepath, report_files_dirs)

        report_files = {}
        for report_files_dir in report_files_dirs:
            report_files.update(self._read_report_from_dir(report_files_dir=report_files_dir))

        return self._handle_include_for_each_report(report_files=report_files)

    def _read_report_from_dir(
        self,
        report_files_dir: str,
    ) -> Dict[str, JSON]:

        report_files = {}
        files = self.file_system.list_entries(path=report_files_dir)
        for report_file in files:
            if self.file_system.is_dir(report_file):
                report_files.update(self._read_report_from_dir(report_files_dir=report_file))
                continue

            base_name = os.path.basename(report_file)
            base_name = base_name.upper()

            report_name, extension = base_name.split(".")
            if extension != "JSON":
                continue

            file_content = self.file_system.read_file_content(path=report_file)
            report_files[report_name] = json.loads(file_content.decode())

        return report_files

    def _handle_include_for_each_report(
        self,
        report_files: Dict[str, List[JSON]],
    ) -> Dict[str, List[JSON]]:

        complete_report_files = {}
        for report_name, report_file in report_files.items():
            complete_report_files[report_name] = self._handle_include_line(
                report_name=report_name,
                report=report_file,
                report_files=report_files,
            )

        return complete_report_files

    def _handle_include_line(
        self,
        report_name: str,
        report: List[JSON],
        report_files: Dict[str, List[JSON]],
        already_processed: Optional[List[str]] = None,
        condition_to_include: Optional[Union[str, List[str]]] = None,
    ) -> List[JSON]:

        if already_processed is None:
            already_processed = []

        if report_name in already_processed:
            raise IncludeCycle(cycle=list(already_processed) + [report_name])

        complete_lines = []
        for report_line in report:
            if condition_to_include:
                report_line["condition"] = self._get_propagated_condition(
                    condition_to_include=condition_to_include,
                    report_line=report_line,
                )

            include_report_name = report_line.get("include")
            if include_report_name is None:
                complete_lines.append(report_line)
            else:
                include_report_name = include_report_name.upper()
                already_processed.append(report_name)
                report_to_include = self._handle_include_line(
                    report_name=include_report_name,
                    report=pickle_deep_copy(obj=report_files[include_report_name]),
                    report_files=report_files,
                    already_processed=already_processed,
                    condition_to_include=report_line.get("condition"),
                )
                already_processed = []
                complete_lines.extend(report_to_include)

        return complete_lines

    @staticmethod
    def _get_propagated_condition(
        condition_to_include: Union[str, List[str]],
        report_line: JSON,
    ) -> List[str]:

        condition = report_line.get("condition")

        if isinstance(condition_to_include, str):
            condition_to_include = [condition_to_include]

        if not condition:
            return condition_to_include

        if isinstance(report_line["condition"], str):
            condition = [condition]

        condition.extend(condition_to_include)

        return list(set(condition))

    def _read_images_path(
        self,
    ) -> Dict[str, str]:

        images_path = self.reader.find_value(
            path="nReports.ImagesPath",
            default=self._default_images_path,
        )
        if not self.file_system.is_dir(images_path):
            return {}

        images_by_path = {}
        for file in self.file_system.list_entries(images_path):
            if not self.file_system.is_file(file):
                continue

            file_name_no_extension, _ = os.path.splitext(os.path.basename(file))

            images_by_path[file_name_no_extension] = file

        return images_by_path

    def _read_constants(
        self,
    ) -> Dict[str, str]:

        constants_group = self.reader.find_group(path="nReports.Constants")

        constants: Dict[str, str] = {}
        for key in constants_group.keys:
            value = key.value()
            if not value:
                continue

            constants[key.name] = value

        return constants



"""Provide basic warnings used by setuptools modules.

Using custom classes (other than ``UserWarning``) allow users to set
``PYTHONWARNINGS`` filters to run tests and prepare for upcoming changes in
setuptools.
"""

from __future__ import annotations

import os
import warnings
from datetime import date
from inspect import cleandoc
from textwrap import indent
from typing import Tuple

_DueDate = Tuple[int, int, int]  # time tuple
_INDENT = 8 * " "
_TEMPLATE = f"""{80 * '*'}\n{{details}}\n{80 * '*'}"""


class SetuptoolsWarning(UserWarning):
    """Base class in ``setuptools`` warning hierarchy."""

    @classmethod
    def emit(
        cls,
        summary: str | None = None,
        details: str | None = None,
        due_date: _DueDate | None = None,
        see_docs: str | None = None,
        see_url: str | None = None,
        stacklevel: int = 2,
        **kwargs,
    ):
        """Private: reserved for ``setuptools`` internal use only"""
        # Default values:
        summary_ = summary or getattr(cls, "_SUMMARY", None) or ""
        details_ = details or getattr(cls, "_DETAILS", None) or ""
        due_date = due_date or getattr(cls, "_DUE_DATE", None)
        docs_ref = see_docs or getattr(cls, "_SEE_DOCS", None)
        docs_url = docs_ref and f"https://setuptools.pypa.io/en/latest/{docs_ref}"
        see_url = see_url or getattr(cls, "_SEE_URL", None)
        due = date(*due_date) if due_date else None

        text = cls._format(summary_, details_, due, see_url or docs_url, kwargs)
        if due and due < date.today() and _should_enforce():
            raise cls(text)
        warnings.warn(text, cls, stacklevel=stacklevel + 1)

    @classmethod
    def _format(
        cls,
        summary: str,
        details: str,
        due_date: date | None = None,
        see_url: str | None = None,
        format_args: dict | None = None,
    ):
        """Private: reserved for ``setuptools`` internal use only"""
        today = date.today()
        summary = cleandoc(summary).format_map(format_args or {})
        possible_parts = [
            cleandoc(details).format_map(format_args or {}),
            (
                f"\nBy {due_date:%Y-%b-%d}, you need to update your project and remove "
                "deprecated calls\nor your builds will no longer be supported."
                if due_date and due_date > today
                else None
            ),
            (
                "\nThis deprecation is overdue, please update your project and remove "
                "deprecated\ncalls to avoid build errors in the future."
                if due_date and due_date < today
                else None
            ),
            (f"\nSee {see_url} for details." if see_url else None),
        ]
        parts = [x for x in possible_parts if x]
        if parts:
            body = indent(_TEMPLATE.format(details="\n".join(parts)), _INDENT)
            return "\n".join([summary, "!!\n", body, "\n!!"])
        return summary


class InformationOnly(SetuptoolsWarning):
    """Currently there is no clear way of displaying messages to the users
    that use the setuptools backend directly via ``pip``.
    The only thing that might work is a warning, although it is not the
    most appropriate tool for the job...

    See pypa/packaging-problems#558.
    """


class SetuptoolsDeprecationWarning(SetuptoolsWarning):
    """
    Base class for warning deprecations in ``setuptools``

    This class is not derived from ``DeprecationWarning``, and as such is
    visible by default.
    """


def _should_enforce():
    enforce = os.getenv("SETUPTOOLS_ENFORCE_DEPRECATION", "false").lower()
    return enforce in ("true", "on", "ok", "1")



from logging import Logger
from xml.etree import ElementTree as eTree

from cfgtools import Configuration
from helper import remove_accents
from pos_model import (
    Fee,
    Order,
    SaleItem,
)
from typing import (
    List,
    Optional,
)

from common import (
    CalculateTaxesItemCommon,
    FiscalParameterController,
    build_item_taxes,
)
from models import FeeDetails

_FEE_DETAILS_CACHE = []    # type: List[FeeDetails]


def get_order_fees(
    order_xml,  # type: eTree.Element
):
    # type: (...) -> List[Fee]

    order_fees = []
    for order_fee in order_xml.findall(".//OrderFees/Fee"):
        fee = Fee(
            code=order_fee.get("code"),
            amount=float(order_fee.get("calculatedAmount")),
        )
        order_fees.append(fee)

    return order_fees


def get_fee_details(
    fee_code,   # type: str
):
    # type: (...) -> Optional[FeeDetails]

    for fee in _FEE_DETAILS_CACHE:
        if fee.code == fee_code:
            return fee

    return None


class FeeService(object):

    def __init__(
        self,                           # type: FeeService
        loader_config,                  # type: Configuration
        logger,                         # type: Logger
        calculate_taxes_common,         # type: CalculateTaxesItemCommon
        fiscal_parameter_controller,    # type: FiscalParameterController
    ):
        # type: (...) -> None

        self.loader_config = loader_config
        self.logger = logger
        self.calculate_taxes_common = calculate_taxes_common
        self.fiscal_parameter_controller = fiscal_parameter_controller

        self._read_fees_details()

    def get_fee_sale_items(
        self,   # type: FeeService
        order,  # type: Order
    ):
        # type: (...) -> List[SaleItem]

        fee_sale_items = []

        for fee in order.fees:
            fee_item = self._build_fee_item(fee=fee)
            if fee_item:
                fee_sale_items.append(fee_item)

        return fee_sale_items

    def _build_fee_item(
        self,   # type: FeeService
        fee,    # type: Fee
    ):
        # type: (...) -> Optional[SaleItem]

        fee_details = get_fee_details(fee_code=fee.code)
        if not fee_details:
            return None

        sale_item = SaleItem()
        sale_item.part_code = fee_details.product_code
        sale_item.product_name = fee_details.product_name.upper()
        sale_item.quantity = 1
        sale_item.unit_price = sale_item.item_price = fee.amount
        sale_item.taxes = build_item_taxes(
            sale_item=sale_item,
            fiscal_parameter_controller=self.fiscal_parameter_controller,
            calculate_taxes_common=self.calculate_taxes_common,
        )

        return sale_item

    def _read_fees_details(
        self,   # type: FeeService
    ):
        # type: (...) -> None

        global _FEE_DETAILS_CACHE

        main_group = self.loader_config.find_group(path="FiscalWrapper.FeesFiscalParameters")
        if not main_group:
            return

        for group in main_group.groups:

            fee_details = FeeDetails(code=str(group.name))

            for key in group.keys:
                key_name = str(key.name)
                if key_name == "PRODUCT_CODE":
                    fee_details.product_code = str(key.values[0])
                elif key_name == "PRODUCT_NAME":
                    fee_details.product_name = str(remove_accents(key.values[0]))

            _FEE_DETAILS_CACHE.append(fee_details)


def order_xml_has_valid_fee(
    order_xml,  # type: eTree.Element
):
    # type: (...) -> bool

    order_fees = get_order_fees(order_xml=order_xml)
    if not order_fees:
        return False

    for order_fee in order_fees:
        if not get_fee_details(fee_code=order_fee.code):
            return False

    return True



from production.box import DisplayTimeFilterBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class DisplayTimeFilterBoxBuilder(BoxBuilder):
    def __init__(self, loggers):
        super(DisplayTimeFilterBoxBuilder, self).__init__(loggers)

    def build(self, config):
        # type: (BoxConfiguration) -> DisplayTimeFilterBox
        return DisplayTimeFilterBox(
            config.name,
            config.sons,
            self.get_extra(config, "ExpirationTime", None),
            self.get_logger(config)
        )



from typing import (
    Dict,
    List,
    Union,
)

from nreports.repository import (
    I18NRepository,
    L10NRepository,
)
from nreports_api.shared import JSON
from ._Line import Line
from .exceptions import (
    InnerListAttributeIsNotAList,
    InnerListMustHaveNameAttr,
)


class InnerListLine(Line):

    def __init__(
        self,
        line: JSON,
    ) -> None:

        super().__init__(line)

        if "name" not in line:
            raise InnerListMustHaveNameAttr()

        if not isinstance(line["innerList"], list):
            raise InnerListAttributeIsNotAList()

        self._inner_list_name = line["name"]
        self._inner_list: List[Line] = []

        for text in line["innerList"]:
            self._inner_list.append(Line.build(line=text))

    def inner_format(
        self,
        data: Dict[str, JSON],
        i18n_repository: I18NRepository,
        l10n_repository: L10NRepository,
    ) -> List[Dict[str, Union[str, Dict[str, str]]]]:

        formatted = []
        for sub_group_data in data[self._inner_list_name]:
            for line in self._inner_list:
                formatted_line = line.format(
                    data=sub_group_data,
                    i18n_repository=i18n_repository,
                    l10n_repository=l10n_repository,
                )
                formatted.extend(formatted_line)

        return formatted

    def get_inner_data_to_build(
        self,
    ) -> Dict[str, JSON]:

        ret: Dict[str, JSON] = {
            "template": {},
            self._inner_list_name: []
        }
        for line in self._inner_list:
            ret["template"].update(line.get_inner_data_to_build())

        return ret

    def set_inner_list(
        self,
        inner_list: List[Line],
    ) -> None:

        self._inner_list = inner_list



from xml.etree import cElementTree as eTree
from sysactions import (
    get_model,
    get_posot,
    OrderTaker,
)
from services.model import OrderCanceledRequest
from externals.repository import AlertsRepository
from services.model import ConfiguredAlert
from services.util import (
    get_configured_alerts_by_type,
    insert_event_into_repository,
    get_order_total,
    get_table_picture,
)
from mwhelper import ModelService


class OrderCanceledInteractor(object):
    def __init__(
        self,
        alerts_repository,  # type: AlertsRepository
    ):
        # type: (...) -> None

        self.alerts_repository = alerts_repository
        self.configured_alerts = self.alerts_repository.get_configured_alerts()

    def process_request(
        self,
        request,  # type: OrderCanceledRequest
    ):
        # type: (...) -> None

        configured_alerts = get_configured_alerts_by_type(
            notify_type="CANCEL",
            configured_alerts=self.configured_alerts,
        )
        if not configured_alerts:
            return

        pos_id = request.pos_id
        order_id = request.order_id
        model = get_model(posid=pos_id)
        pos_ot = get_posot(model=model)
        order_picture = eTree.XML(pos_ot.orderPicture(orderid=order_id)).find("Order")

        if not self._should_notify_order(order_picture=order_picture):
            return

        is_table = ModelService.is_table_service(order_picture)
        fiscalization_date = "CustomOrderProperties/OrderProperty[@key='FISCALIZATION_DATE']"
        has_fiscalization_date = order_picture.find(fiscalization_date) is not None
        state_history = eTree.tostring(order_picture.find("StateHistory"))
        is_fiscalized = "PAID" in state_history and has_fiscalization_date
        if is_table and not is_fiscalized:
            total = self._get_table_total(
                pos_id=pos_id,
                order_picture=order_picture,
                pos_ot=pos_ot,
            )
        else:
            total = get_order_total(order_picture=order_picture)

        order_picture_dict = None
        table_picture_dict = None
        for alert in configured_alerts:
            notify_name = "cancelAmount"
            if is_table:
                notify_name = "tableCancelAmount"
            valid_conditions = self._validate_alert(
                alert=alert,
                has_fiscalization_date=has_fiscalization_date,
                total=total,
                notify_name=notify_name,
                is_table=is_table
            )
            if not valid_conditions:
                continue

            order_picture_dict, table_picture_dict = insert_event_into_repository(
                pos_id=pos_id,
                alerts_repository=self.alerts_repository,
                order_picture=order_picture,
                alert=alert,
                order_picture_dict=order_picture_dict,
                table_picture_dict=table_picture_dict,
            )

    @staticmethod
    def _should_notify_order(
        order_picture,    # type: eTree.Element
    ):
        # type: (...) -> bool

        state_history = eTree.tostring(order_picture.find("StateHistory"))
        was_stored = "STORED" in state_history

        fiscalization_date = "CustomOrderProperties/OrderProperty[@key='FISCALIZATION_DATE']"
        has_fiscalization_date = order_picture.find(fiscalization_date) is not None
        is_fiscalized = "PAID" in state_history and has_fiscalization_date

        is_table = ModelService.is_table_service(order_picture)
        if is_table:
            if was_stored or is_fiscalized:
                return True
        else:
            is_stored_drive_thru = order_picture.get("saleTypeDescr") == "DRIVE_THRU" and was_stored
            if is_stored_drive_thru:
                return True

            return is_fiscalized

        return False

    @staticmethod
    def _validate_alert(
        alert,                   # type: ConfiguredAlert
        has_fiscalization_date,  # type: bool
        total,                   # type: float
        notify_name,             # type: str
        is_table,                # type: bool
    ):
        # type: (...) -> bool

        alert_conditions = alert.alert_json.get("conditions")
        notify_cancel = notify_name in alert_conditions
        reached_cancel_amount = total >= float(alert_conditions.get(notify_name))
        if notify_cancel and not reached_cancel_amount:
            return False

        alert_after_fiscalization = "onlyAfterFiscalization" in alert_conditions
        if alert_after_fiscalization:
            only_after_fiscalization = alert_conditions.get("onlyAfterFiscalization") in ("True", True)
            if only_after_fiscalization and not has_fiscalization_date:
                return False
        elif not has_fiscalization_date and not is_table:
            return False

        return True

    @staticmethod
    def _get_table_total(
        pos_id,             # type: str
        order_picture,      # type: eTree.XML
        pos_ot,             # type: OrderTaker
    ):
        # type: (...) -> float

        table_id = order_picture.find("CustomOrderProperties/OrderProperty[@key='TABLE_ID']").get("value")
        table_picture = eTree.XML(
            get_table_picture(
                pos_id=pos_id,
                table_id=table_id,
            )
        )

        service_orders = table_picture.findall("ServiceOrders/ServiceOrder")
        table_total = 0
        for order in service_orders:
            state = order.get("state")
            order_id = order.get("orderId")
            order_picture = eTree.XML(pos_ot.orderPicture(orderid=order_id)).find("Order")
            state_history = eTree.tostring(order_picture.find("StateHistory"))
            if state != "VOIDED" or "STORED" not in state_history:
                continue
            table_total += get_order_total(order_picture=order_picture)
        return table_total



# -*- coding: utf-8 -*-

# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: google/api/routing.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x18google/api/routing.proto\x12\ngoogle.api\x1a google/protobuf/descriptor.proto"G\n\x0bRoutingRule\x12\x38\n\x12routing_parameters\x18\x02 \x03(\x0b\x32\x1c.google.api.RoutingParameter"8\n\x10RoutingParameter\x12\r\n\x05\x66ield\x18\x01 \x01(\t\x12\x15\n\rpath_template\x18\x02 \x01(\t:K\n\x07routing\x12\x1e.google.protobuf.MethodOptions\x18\xb1\xca\xbc" \x01(\x0b\x32\x17.google.api.RoutingRuleBj\n\x0e\x63om.google.apiB\x0cRoutingProtoP\x01ZAgoogle.golang.org/genproto/googleapis/api/annotations;annotations\xa2\x02\x04GAPIb\x06proto3'
)


ROUTING_FIELD_NUMBER = 72295729
routing = DESCRIPTOR.extensions_by_name["routing"]

_ROUTINGRULE = DESCRIPTOR.message_types_by_name["RoutingRule"]
_ROUTINGPARAMETER = DESCRIPTOR.message_types_by_name["RoutingParameter"]
RoutingRule = _reflection.GeneratedProtocolMessageType(
    "RoutingRule",
    (_message.Message,),
    {
        "DESCRIPTOR": _ROUTINGRULE,
        "__module__": "google.api.routing_pb2"
        # @@protoc_insertion_point(class_scope:google.api.RoutingRule)
    },
)
_sym_db.RegisterMessage(RoutingRule)

RoutingParameter = _reflection.GeneratedProtocolMessageType(
    "RoutingParameter",
    (_message.Message,),
    {
        "DESCRIPTOR": _ROUTINGPARAMETER,
        "__module__": "google.api.routing_pb2"
        # @@protoc_insertion_point(class_scope:google.api.RoutingParameter)
    },
)
_sym_db.RegisterMessage(RoutingParameter)

if _descriptor._USE_C_DESCRIPTORS == False:
    google_dot_protobuf_dot_descriptor__pb2.MethodOptions.RegisterExtension(routing)

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"\n\016com.google.apiB\014RoutingProtoP\001ZAgoogle.golang.org/genproto/googleapis/api/annotations;annotations\242\002\004GAPI"
    _ROUTINGRULE._serialized_start = 74
    _ROUTINGRULE._serialized_end = 145
    _ROUTINGPARAMETER._serialized_start = 147
    _ROUTINGPARAMETER._serialized_end = 203
# @@protoc_insertion_point(module_scope)



# tfn.py - functions for handling Australian Tax File Numbers (TFNs)
#
# Copyright (C) 2016 Vincent Bastos
# Copyright (C) 2016 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""TFN (Australian Tax File Number).

The Tax File Number (TFN) is issued by the Australian Taxation Office (ATO)
to taxpaying individuals and organisations. A business has both a TFN and an
Australian Business Number (ABN).

The number consists of 8 (older numbers) or 9 digits and includes a check
digit but otherwise without structure.

More information:

* https://en.wikipedia.org/wiki/Tax_file_number
* https://www.ato.gov.au/Individuals/Tax-file-number/

>>> validate('123 456 782')
'123456782'
>>> validate('999 999 999')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> format('123456782')
'123 456 782'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' ').strip()


def checksum(number):
    """Calculate the checksum."""
    weights = (1, 4, 3, 7, 5, 8, 6, 9, 10)
    return sum(w * int(n) for w, n in zip(weights, number)) % 11


def validate(number):
    """Check if the number is a valid TFN. This checks the length, formatting
    and check digit."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) not in (8, 9):
        raise InvalidLength()
    if checksum(number) != 0:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid TFN."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return ' '.join((number[0:3], number[3:6], number[6:]))



import unicodedata
import sys
from configparser import ConfigParser

from .compat import py39
from .warnings import SetuptoolsDeprecationWarning


# HFS Plus uses decomposed UTF-8
def decompose(path):
    if isinstance(path, str):
        return unicodedata.normalize('NFD', path)
    try:
        path = path.decode('utf-8')
        path = unicodedata.normalize('NFD', path)
        path = path.encode('utf-8')
    except UnicodeError:
        pass  # Not UTF-8
    return path


def filesys_decode(path):
    """
    Ensure that the given path is decoded,
    ``None`` when no expected encoding works
    """

    if isinstance(path, str):
        return path

    fs_enc = sys.getfilesystemencoding() or 'utf-8'
    candidates = fs_enc, 'utf-8'

    for enc in candidates:
        try:
            return path.decode(enc)
        except UnicodeDecodeError:
            continue

    return None


def try_encode(string, enc):
    "turn unicode encoding into a functional routine"
    try:
        return string.encode(enc)
    except UnicodeEncodeError:
        return None


def _read_utf8_with_fallback(file: str, fallback_encoding=py39.LOCALE_ENCODING) -> str:
    """
    First try to read the file with UTF-8, if there is an error fallback to a
    different encoding ("locale" by default). Returns the content of the file.
    Also useful when reading files that might have been produced by an older version of
    setuptools.
    """
    try:
        with open(file, "r", encoding="utf-8") as f:
            return f.read()
    except UnicodeDecodeError:  # pragma: no cover
        _Utf8EncodingNeeded.emit(file=file, fallback_encoding=fallback_encoding)
        with open(file, "r", encoding=fallback_encoding) as f:
            return f.read()


def _cfg_read_utf8_with_fallback(
    cfg: ConfigParser, file: str, fallback_encoding=py39.LOCALE_ENCODING
) -> None:
    """Same idea as :func:`_read_utf8_with_fallback`, but for the
    :meth:`ConfigParser.read` method.

    This method may call ``cfg.clear()``.
    """
    try:
        cfg.read(file, encoding="utf-8")
    except UnicodeDecodeError:  # pragma: no cover
        _Utf8EncodingNeeded.emit(file=file, fallback_encoding=fallback_encoding)
        cfg.clear()
        cfg.read(file, encoding=fallback_encoding)


class _Utf8EncodingNeeded(SetuptoolsDeprecationWarning):
    _SUMMARY = """
    `encoding="utf-8"` fails with {file!r}, trying `encoding={fallback_encoding!r}`.
    """

    _DETAILS = """
    Fallback behaviour for UTF-8 is considered **deprecated** and future versions of
    `setuptools` may not implement it.

    Please encode {file!r} with "utf-8" to ensure future builds will succeed.

    If this file was produced by `setuptools` itself, cleaning up the cached files
    and re-building/re-installing the package with a newer version of `setuptools`
    (e.g. by updating `build-system.requires` in its `pyproject.toml`)
    might solve the problem.
    """
    # TODO: Add a deadline?
    #       Will we be able to remove this?
    #       The question comes to mind mainly because of sdists that have been produced
    #       by old versions of setuptools and published to PyPI...



# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

# Copyright (C) 2001-2017 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose with or without fee is hereby granted,
# provided that the above copyright notice and this permission notice
# appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""DNS rdata."""

from io import BytesIO
import base64
import binascii

import dns.exception
import dns.name
import dns.rdataclass
import dns.rdatatype
import dns.tokenizer
import dns.wiredata
from ._compat import xrange, string_types, text_type

try:
    import threading as _threading
except ImportError:
    import dummy_threading as _threading

_hex_chunksize = 32


def _hexify(data, chunksize=_hex_chunksize):
    """Convert a binary string into its hex encoding, broken up into chunks
    of chunksize characters separated by a space.
    """

    line = binascii.hexlify(data)
    return b' '.join([line[i:i + chunksize]
                      for i
                      in range(0, len(line), chunksize)]).decode()

_base64_chunksize = 32


def _base64ify(data, chunksize=_base64_chunksize):
    """Convert a binary string into its base64 encoding, broken up into chunks
    of chunksize characters separated by a space.
    """

    line = base64.b64encode(data)
    return b' '.join([line[i:i + chunksize]
                      for i
                      in range(0, len(line), chunksize)]).decode()

__escaped = bytearray(b'"\\')

def _escapify(qstring):
    """Escape the characters in a quoted string which need it."""

    if isinstance(qstring, text_type):
        qstring = qstring.encode()
    if not isinstance(qstring, bytearray):
        qstring = bytearray(qstring)

    text = ''
    for c in qstring:
        if c in __escaped:
            text += '\\' + chr(c)
        elif c >= 0x20 and c < 0x7F:
            text += chr(c)
        else:
            text += '\\%03d' % c
    return text


def _truncate_bitmap(what):
    """Determine the index of greatest byte that isn't all zeros, and
    return the bitmap that contains all the bytes less than that index.
    """

    for i in xrange(len(what) - 1, -1, -1):
        if what[i] != 0:
            return what[0: i + 1]
    return what[0:1]


class Rdata(object):
    """Base class for all DNS rdata types."""

    __slots__ = ['rdclass', 'rdtype']

    def __init__(self, rdclass, rdtype):
        """Initialize an rdata.

        *rdclass*, an ``int`` is the rdataclass of the Rdata.
        *rdtype*, an ``int`` is the rdatatype of the Rdata.
        """

        self.rdclass = rdclass
        self.rdtype = rdtype

    def covers(self):
        """Return the type a Rdata covers.

        DNS SIG/RRSIG rdatas apply to a specific type; this type is
        returned by the covers() function.  If the rdata type is not
        SIG or RRSIG, dns.rdatatype.NONE is returned.  This is useful when
        creating rdatasets, allowing the rdataset to contain only RRSIGs
        of a particular type, e.g. RRSIG(NS).

        Returns an ``int``.
        """

        return dns.rdatatype.NONE

    def extended_rdatatype(self):
        """Return a 32-bit type value, the least significant 16 bits of
        which are the ordinary DNS type, and the upper 16 bits of which are
        the "covered" type, if any.

        Returns an ``int``.
        """

        return self.covers() << 16 | self.rdtype

    def to_text(self, origin=None, relativize=True, **kw):
        """Convert an rdata to text format.

        Returns a ``text``.
        """

        raise NotImplementedError

    def to_wire(self, file, compress=None, origin=None):
        """Convert an rdata to wire format.

        Returns a ``binary``.
        """

        raise NotImplementedError

    def to_digestable(self, origin=None):
        """Convert rdata to a format suitable for digesting in hashes.  This
        is also the DNSSEC canonical form.

        Returns a ``binary``.
        """

        f = BytesIO()
        self.to_wire(f, None, origin)
        return f.getvalue()

    def validate(self):
        """Check that the current contents of the rdata's fields are
        valid.

        If you change an rdata by assigning to its fields,
        it is a good idea to call validate() when you are done making
        changes.

        Raises various exceptions if there are problems.

        Returns ``None``.
        """

        dns.rdata.from_text(self.rdclass, self.rdtype, self.to_text())

    def __repr__(self):
        covers = self.covers()
        if covers == dns.rdatatype.NONE:
            ctext = ''
        else:
            ctext = '(' + dns.rdatatype.to_text(covers) + ')'
        return '<DNS ' + dns.rdataclass.to_text(self.rdclass) + ' ' + \
               dns.rdatatype.to_text(self.rdtype) + ctext + ' rdata: ' + \
               str(self) + '>'

    def __str__(self):
        return self.to_text()

    def _cmp(self, other):
        """Compare an rdata with another rdata of the same rdtype and
        rdclass.

        Return < 0 if self < other in the DNSSEC ordering, 0 if self
        == other, and > 0 if self > other.

        """
        our = self.to_digestable(dns.name.root)
        their = other.to_digestable(dns.name.root)
        if our == their:
            return 0
        elif our > their:
            return 1
        else:
            return -1

    def __eq__(self, other):
        if not isinstance(other, Rdata):
            return False
        if self.rdclass != other.rdclass or self.rdtype != other.rdtype:
            return False
        return self._cmp(other) == 0

    def __ne__(self, other):
        if not isinstance(other, Rdata):
            return True
        if self.rdclass != other.rdclass or self.rdtype != other.rdtype:
            return True
        return self._cmp(other) != 0

    def __lt__(self, other):
        if not isinstance(other, Rdata) or \
                self.rdclass != other.rdclass or self.rdtype != other.rdtype:

            return NotImplemented
        return self._cmp(other) < 0

    def __le__(self, other):
        if not isinstance(other, Rdata) or \
                self.rdclass != other.rdclass or self.rdtype != other.rdtype:
            return NotImplemented
        return self._cmp(other) <= 0

    def __ge__(self, other):
        if not isinstance(other, Rdata) or \
                self.rdclass != other.rdclass or self.rdtype != other.rdtype:
            return NotImplemented
        return self._cmp(other) >= 0

    def __gt__(self, other):
        if not isinstance(other, Rdata) or \
                self.rdclass != other.rdclass or self.rdtype != other.rdtype:
            return NotImplemented
        return self._cmp(other) > 0

    def __hash__(self):
        return hash(self.to_digestable(dns.name.root))

    @classmethod
    def from_text(cls, rdclass, rdtype, tok, origin=None, relativize=True):
        raise NotImplementedError

    @classmethod
    def from_wire(cls, rdclass, rdtype, wire, current, rdlen, origin=None):
        raise NotImplementedError

    def choose_relativity(self, origin=None, relativize=True):
        """Convert any domain names in the rdata to the specified
        relativization.
        """

class GenericRdata(Rdata):

    """Generic Rdata Class

    This class is used for rdata types for which we have no better
    implementation.  It implements the DNS "unknown RRs" scheme.
    """

    __slots__ = ['data']

    def __init__(self, rdclass, rdtype, data):
        super(GenericRdata, self).__init__(rdclass, rdtype)
        self.data = data

    def to_text(self, origin=None, relativize=True, **kw):
        return r'\# %d ' % len(self.data) + _hexify(self.data)

    @classmethod
    def from_text(cls, rdclass, rdtype, tok, origin=None, relativize=True):
        token = tok.get()
        if not token.is_identifier() or token.value != r'\#':
            raise dns.exception.SyntaxError(
                r'generic rdata does not start with \#')
        length = tok.get_int()
        chunks = []
        while 1:
            token = tok.get()
            if token.is_eol_or_eof():
                break
            chunks.append(token.value.encode())
        hex = b''.join(chunks)
        data = binascii.unhexlify(hex)
        if len(data) != length:
            raise dns.exception.SyntaxError(
                'generic rdata hex data has wrong length')
        return cls(rdclass, rdtype, data)

    def to_wire(self, file, compress=None, origin=None):
        file.write(self.data)

    @classmethod
    def from_wire(cls, rdclass, rdtype, wire, current, rdlen, origin=None):
        return cls(rdclass, rdtype, wire[current: current + rdlen])

_rdata_modules = {}
_module_prefix = 'dns.rdtypes'
_import_lock = _threading.Lock()

def get_rdata_class(rdclass, rdtype):

    def import_module(name):
        with _import_lock:
            mod = __import__(name)
            components = name.split('.')
            for comp in components[1:]:
                mod = getattr(mod, comp)
            return mod

    mod = _rdata_modules.get((rdclass, rdtype))
    rdclass_text = dns.rdataclass.to_text(rdclass)
    rdtype_text = dns.rdatatype.to_text(rdtype)
    rdtype_text = rdtype_text.replace('-', '_')
    if not mod:
        mod = _rdata_modules.get((dns.rdatatype.ANY, rdtype))
        if not mod:
            try:
                mod = import_module('.'.join([_module_prefix,
                                              rdclass_text, rdtype_text]))
                _rdata_modules[(rdclass, rdtype)] = mod
            except ImportError:
                try:
                    mod = import_module('.'.join([_module_prefix,
                                                  'ANY', rdtype_text]))
                    _rdata_modules[(dns.rdataclass.ANY, rdtype)] = mod
                except ImportError:
                    mod = None
    if mod:
        cls = getattr(mod, rdtype_text)
    else:
        cls = GenericRdata
    return cls


def from_text(rdclass, rdtype, tok, origin=None, relativize=True):
    """Build an rdata object from text format.

    This function attempts to dynamically load a class which
    implements the specified rdata class and type.  If there is no
    class-and-type-specific implementation, the GenericRdata class
    is used.

    Once a class is chosen, its from_text() class method is called
    with the parameters to this function.

    If *tok* is a ``text``, then a tokenizer is created and the string
    is used as its input.

    *rdclass*, an ``int``, the rdataclass.

    *rdtype*, an ``int``, the rdatatype.

    *tok*, a ``dns.tokenizer.Tokenizer`` or a ``text``.

    *origin*, a ``dns.name.Name`` (or ``None``), the
    origin to use for relative names.

    *relativize*, a ``bool``.  If true, name will be relativized to
    the specified origin.

    Returns an instance of the chosen Rdata subclass.
    """

    if isinstance(tok, string_types):
        tok = dns.tokenizer.Tokenizer(tok)
    cls = get_rdata_class(rdclass, rdtype)
    if cls != GenericRdata:
        # peek at first token
        token = tok.get()
        tok.unget(token)
        if token.is_identifier() and \
           token.value == r'\#':
            #
            # Known type using the generic syntax.  Extract the
            # wire form from the generic syntax, and then run
            # from_wire on it.
            #
            rdata = GenericRdata.from_text(rdclass, rdtype, tok, origin,
                                           relativize)
            return from_wire(rdclass, rdtype, rdata.data, 0, len(rdata.data),
                             origin)
    return cls.from_text(rdclass, rdtype, tok, origin, relativize)


def from_wire(rdclass, rdtype, wire, current, rdlen, origin=None):
    """Build an rdata object from wire format

    This function attempts to dynamically load a class which
    implements the specified rdata class and type.  If there is no
    class-and-type-specific implementation, the GenericRdata class
    is used.

    Once a class is chosen, its from_wire() class method is called
    with the parameters to this function.

    *rdclass*, an ``int``, the rdataclass.

    *rdtype*, an ``int``, the rdatatype.

    *wire*, a ``binary``, the wire-format message.

    *current*, an ``int``, the offset in wire of the beginning of
    the rdata.

    *rdlen*, an ``int``, the length of the wire-format rdata

    *origin*, a ``dns.name.Name`` (or ``None``).  If not ``None``,
    then names will be relativized to this origin.

    Returns an instance of the chosen Rdata subclass.
    """

    wire = dns.wiredata.maybe_wrap(wire)
    cls = get_rdata_class(rdclass, rdtype)
    return cls.from_wire(rdclass, rdtype, wire, current, rdlen, origin)


class RdatatypeExists(dns.exception.DNSException):
    """DNS rdatatype already exists."""
    supp_kwargs = {'rdclass', 'rdtype'}
    fmt = "The rdata type with class {rdclass} and rdtype {rdtype} " + \
        "already exists."


def register_type(implementation, rdtype, rdtype_text, is_singleton=False,
                  rdclass=dns.rdataclass.IN):
    """Dynamically register a module to handle an rdatatype.

    *implementation*, a module implementing the type in the usual dnspython
    way.

    *rdtype*, an ``int``, the rdatatype to register.

    *rdtype_text*, a ``text``, the textual form of the rdatatype.

    *is_singleton*, a ``bool``, indicating if the type is a singleton (i.e.
    RRsets of the type can have only one member.)

    *rdclass*, the rdataclass of the type, or ``dns.rdataclass.ANY`` if
    it applies to all classes.
    """

    existing_cls = get_rdata_class(rdclass, rdtype)
    if existing_cls != GenericRdata:
        raise RdatatypeExists(rdclass=rdclass, rdtype=rdtype)
    _rdata_modules[(rdclass, rdtype)] = implementation
    dns.rdatatype.register_type(rdtype, rdtype_text, is_singleton)
